from pysqream.casting import *
from pysqream.globals import WIN, BUFFER_SIZE, ROWS_PER_FLUSH, VARCHAR_ENCODING, type_to_letter, ARROW, buf_maps, buf_views
from pysqream.utils import NotSupportedError, ProgrammingError, InternalError, IntegrityError, OperationalError, DataError, \
    DatabaseError, InterfaceError, Warning, Error

from pysqream.logger import *
from decimal import Decimal
import traceback
import array
from struct import pack_into, error as struct_error
import multiprocessing as mp
import numpy as np
from mmap import mmap
from functools import reduce
import functools
import operator


def init_lock(l):

    """To pass a lock to mp.Pool()"""

    global lock
    lock = l


class ColumnBuffer:
    ''' Buffer holding packed columns to be sent to SQream '''

    def __init__(self, size=BUFFER_SIZE):
        global buf_maps, buf_views

    def clear(self):
        if buf_maps:
            [buf_map.close() for buf_map in buf_maps[0]]

    def init_buffers(self, col_sizes, col_nul):
        if not WIN:
            try:
                self.pool.close()
                self.pool.join()
            except Exception as e:
                pass

            l = mp.Lock()
            self.pool = mp.Pool(initializer=init_lock, initargs=(l,))
        self.clear()
        buf_maps = [mmap(-1, ((1 if col_nul else 0) + (size if size != 0 else 104)) * ROWS_PER_FLUSH) for size in
                    col_sizes]
        buf_views = [memoryview(buf_map) for buf_map in buf_maps]

    def pack_columns(self, cols, capacity, col_types, col_sizes, col_nul, col_tvc, col_scales):
        ''' Packs the buffer starting a given index with the column.
            Returns number of bytes packed '''

        pool_params = zip(cols, range(len(col_types)), col_types,
                          col_sizes, col_nul, col_tvc, col_scales)
        if WIN:
            packed_cols = []
            for param_tup in pool_params:
                packed_cols.append(_pack_column(param_tup))

        else:
            # self.pool = mp.Pool()
            # To use multiprocess type packing, we call a top level function with a single tuple parameter
            try:
                packed_cols = self.pool.map(_pack_column, pool_params, chunksize=2)  # buf_end_indices
            except Exception as e:
                printdbg("Original error from pool.map: ", e)
                if logger.isEnabledFor(logging.ERROR):
                    logger.error("Original error from pool.map: ", e)
                log_and_raise(ProgrammingError,
                              "Error packing columns. Check that all types match the respective column types")

        return list(packed_cols)

    def close(self):
        self.clear()
        try:
            self.pool.close()
            self.pool.join()
        except Exception as e:
            # print (f'testing pool closing, got: {e}')
            pass  # no pool was initiated


## A top level packing function for Python's MP compatibility
def _pack_column(col_tup, return_actual_data=True):
    ''' Packs the buffer starting a given index with the column.
        Returns number of bytes packed '''

    global CYTHON
    col, col_idx, col_type, size, nullable, tvc, scale = col_tup
    col = list(col)
    capacity = len(col)
    buf_idx = 0
    buf_map = mmap(-1, ((1 if nullable else 0) + (size if size != 0 else 104)) * ROWS_PER_FLUSH)
    buf_view = memoryview(buf_map)

    def pack_exception(e):
        ''' Allowing to return traceback info from parent process when using mp.Pool on _pack_column
            [add link]
        '''

        e.traceback = traceback.format_exc()
        error_msg = f'Trying to insert unsuitable types to column number {col_idx + 1} of type {col_type}'
        with lock:
            logger.error(error_msg, exc_info=True)
        raise ProgrammingError(error_msg)

    # Numpy array for column
    if ARROW and isinstance(col, np.ndarray):
        # Pack null column if applicable
        if nullable:
            pack_into(f'{capacity}b', buf_view, buf_idx,
                      *[1 if item in (np.nan, b'') else 0 for item in col])
            buf_idx += capacity

            # Replace Nones with appropriate placeholder
            if 'S' in repr(col.dtype):  # already b''?
                pass
            elif 'U' in repr(col.dtype):
                pass
            else:
                # Swap out the nans
                col[col == np.nan] = 0

        # Pack nvarchar length column if applicable
        if tvc:
            buf_map.seek(buf_idx)
            lengths_as_bytes = np.vectorize(len)(col).astype('int32').tobytes()
            buf_map.write(lengths_as_bytes)
            buf_idx += len(lengths_as_bytes)

        # Pack the actual data
        if 'U' in repr(col.dtype):
            packed_strings = ''.join(col).encode('utf8')
            buf_map.seek(buf_idx)
            buf_map.write(packed_strings)
            buf_idx += len(packed_strings)
            # print(f'unicode strings: {packed_strings}')
        else:
            packed_np = col.tobytes()
            buf_map.seek(buf_idx)
            buf_map.write(packed_np)
            buf_idx += len(packed_np)

        return buf_map[0:buf_idx] if return_actual_data else (0, buf_idx)

    # Pack null column if applicable
    type_code = type_to_letter[col_type]

    # Pack null column and replace None with appropriate placeholder
    col_placeholder = {
        'ftBool': 0,
        'ftUByte': 0,
        'ftShort': 0,
        'ftInt': 0,
        'ftLong': 0,
        'ftFloat': 0,
        'ftDouble': 0,
        'ftDate': None,  # updated separately
        'ftDateTime': None,  # updated separately
        'ftVarchar': ''.ljust(size, ' '),
        'ftBlob': '',
        'ftNumeric': 0
    }

    if nullable:
        idx = -1
        while True:
            try:
                idx = col.index(None, idx + 1)
            except ValueError:
                break
            else:
                buf_map.seek(buf_idx + idx)
                buf_map.write(b'\x01')
                col[idx] = col_placeholder[col_type]
        # buf_map.seek(buf_idx)
        # buf_map.write(nulls)
        buf_idx += capacity

    # If a text column, replace and pack in adavnce to see if the buffer is sufficient
    if col_type == 'ftBlob':
        try:
            encoded_col = [strn.encode('utf8') for strn in col]
        except AttributeError as e:  # Non strings will not have .encode()
            pack_exception(e)
        else:
            packed_strings = b''.join(encoded_col)

        needed_buf_size = len(packed_strings) + 5 * capacity

        # Resize the buffer if not enough space for the current strings
        if needed_buf_size > len(buf_map):
            buf_view.release()
            buf_map.resize(needed_buf_size)
            buf_view = memoryview(buf_map)

        # Pack nvarchar length column
        pack_into(f'{capacity}i', buf_view, buf_idx, *[len(string) for string in encoded_col])

        buf_idx += 4 * capacity

    # Replace Nones with appropriate placeholder - this affects the data itself
    if col_type == 'ftBlob':
        pass  # Handled preemptively due to allow possible buffer resizing

    elif col_type == 'ftVarchar':
        try:
            col = (strn.encode(VARCHAR_ENCODING)[:size].ljust(size, b' ') for strn in col)
        except AttributeError as e:  # Non strings will not have .encode()
            pack_exception(e)
        else:
            packed_strings = b''.join(col)

    elif col_type == 'ftDate':
        # date_tuple_to_int(1900, 1, 1) = 693901
        pass
        # '''
        try:
            col = (date_to_int(deit) if deit is not None else 693901 for deit in col)
        except AttributeError as e:  # Non date/times will not have .timetuple()
            pack_exception(e)
        # '''

    elif col_type == 'ftDateTime':
        # datetime_tuple_to_long(1900, 1, 1, 0, 0, 0) = 2980282101661696
        try:
            col = (datetime_to_long(dt) if dt is not None else 2980282101661696 for dt in col)
        except AttributeError as e:
            pack_exception(e)

    elif col_type == 'ftNumeric':
        try:
            col = (decimal_to_sq_numeric(Decimal(num), scale) for num in col)
        except AttributeError as e:
            pack_exception(e)

    elif col_type in ('ftBool', 'ftUByte', 'ftShort', 'ftInt', 'ftLong', 'ftFloat', 'ftDouble'):
        pass
    else:
        error_msg = f'Bad column type passed: {col_type}'
        with lock:
            logger.error(error_msg, exc_info=True)
        raise ProgrammingError(error_msg)

    CYTHON = False
    # Done preceding column handling, pack the actual data
    if col_type in ('ftVarchar', 'ftBlob'):
        buf_map.seek(buf_idx)
        buf_map.write(packed_strings)
        buf_idx += len(packed_strings)
    elif col_type == 'ftNumeric':
        buf_map.seek(buf_idx)
        all = functools.reduce(operator.iconcat, (num.to_bytes(16, byteorder='little', signed=True) for num in col), [])
        buf_map.write(bytearray(all))
        buf_idx += len(all)
    else:
        try:
            if CYTHON:
                buf_map.seek(buf_idx)
                # type_packer[col_type](col, size, buf_map, buf_idx)
            else:
                pack_into(f'{capacity}{type_code}', buf_view, buf_idx, *col)
        except struct_error as e:
            pack_exception(e)

        buf_idx += capacity * size

    return buf_map[0:buf_idx] if return_actual_data else (0, buf_idx)


from datetime import datetime, date, time as t
from decimal import Decimal, getcontext
from math import floor, ceil


def pad_dates(num):
    return ('0' if num < 10 else '') + str(num)


def sq_date_to_py_date(sqream_date, date_convert_func=date):
    if sqream_date is None or sqream_date == 0:
        return None

    year = (10000 * sqream_date + 14780) // 3652425
    intermed_1 = 365 * year + year // 4 - year // 100 + year // 400
    intermed_2 = sqream_date - intermed_1
    if intermed_2 < 0:
        year = year - 1
        intermed_2 = sqream_date - (365 * year + year // 4 - year // 100 +
                                    year // 400)
    intermed_3 = (100 * intermed_2 + 52) // 3060

    year = year + (intermed_3 + 2) // 12
    month = int((intermed_3 + 2) % 12) + 1
    day = int(intermed_2 - (intermed_3 * 306 + 5) // 10 + 1)

    return date_convert_func(year, month, day)


def sq_datetime_to_py_datetime(sqream_datetime, dt_convert_func=datetime):
    ''' Getting the datetime items involves breaking the long into the date int and time it holds
        The date is extracted in the above, while the time is extracted here  '''

    if sqream_datetime is None or sqream_datetime == 0:
        return None

    date_part = sqream_datetime >> 32
    time_part = sqream_datetime & 0xffffffff
    date_part = sq_date_to_py_date(date_part)

    if date_part is None:
        return None

    msec = time_part % 1000
    sec = (time_part // 1000) % 60
    mins = (time_part // 1000 // 60) % 60
    hour = time_part // 1000 // 60 // 60

    return dt_convert_func(date_part.year, date_part.month, date_part.day,
                           hour, mins, sec, msec)


def date_to_int(d: date) -> int:
    year, month, day = d.timetuple()[:3]
    mth: int = (month + 9) % 12
    yr: int = year - mth // 10

    return 365 * yr + yr // 4 - yr // 100 + yr // 400 + (mth * 306 + 5) // 10 + (day - 1)


def datetime_to_long(dt: datetime) -> int:
    ''' self contained to avoid function calling overhead '''

    year, month, day, hour, minute, second = dt.timetuple()[:6]
    msecond = dt.microsecond

    mth: int = (month + 9) % 12
    yr: int = year - mth // 10
    date_int: int = 365 * yr + yr // 4 - yr // 100 + yr // 400 + (
            mth * 306 + 5) // 10 + (day - 1)
    time_int: int = hour * 3600 * 1000 + minute * 60 * 1000 + second * 1000 + msecond // 1000

    return (date_int << 32) + time_int


tenth = Decimal("0.1")
if getcontext().prec < 38:
    getcontext().prec = 38


def sq_numeric_to_decimal(bigint_as_bytes: bytes, scale: int) -> Decimal:
    getcontext().prec = 38
    c = memoryview(bigint_as_bytes).cast('i')
    bigint = ((c[3] << 96) + ((c[2] & 0xffffffff) << 64) + ((c[1] & 0xffffffff) << 32) + (c[0] & 0xffffffff))

    return Decimal(bigint) * (tenth ** scale)


def decimal_to_sq_numeric(dec: Decimal, scale: int) -> int:  # returns bigint
    if getcontext().prec < 38:
        getcontext().prec = 38
    res = dec * (10 ** scale)
    return ceil(res) if res > 0 else floor(res)


def lengths_to_pairs(nvarc_lengths):
    ''' Accumulative sum generator, used for parsing nvarchar columns '''

    idx = new_idx = 0
    for length in nvarc_lengths:
        new_idx += length
        yield idx, new_idx
        idx = new_idx


def numpy_datetime_str_to_tup(numpy_dt):
    ''' '1970-01-01T00:00:00.699148800' '''

    numpy_dt = repr(numpy_dt).split("'")[1]
    date_part, time_part = numpy_dt.split('T')
    year, month, day = date_part.split('-')
    hms, ns = time_part.split('.')
    hour, mins, sec = hms.split(':')
    return year, month, day, hour, mins, sec, ns


def numpy_datetime_str_to_tup2(numpy_dt):
    ''' '1970-01-01T00:00:00.699148800' '''

    ts = (numpy_dt - np.datetime64('1970-01-01T00:00:00Z')) / np.timedelta64(1, 's')
    dt = datetime.utcfromtimestamp(ts)

    return dt.year, dt.month, dt.day



from pysqream.globals import dbg
import logging


logger = logging.getLogger("dbapi_logger")
logger.setLevel(logging.DEBUG)
logger.disabled = True


def printdbg(*debug_print):
    if dbg:
        print(*debug_print)


def start_logging(log_path=None):
    log_path = log_path or '/tmp/sqream_dbapi.log'
    # logging.disable(logging.NOTSET)
    logger.disabled = False
    try:
        handler = logging.FileHandler(log_path)
    except Exception as e:
        raise Exception("Bad log path was given, please verify path is valid and no forbidden characters were used")

    handler.setLevel(logging.DEBUG)
    handler.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
    logger.addHandler(handler)

    return logger


def stop_logging():
    # logging.disable(logging.CRITICAL)
    logger.handlers = []
    logger.disabled = True


def log_and_raise(exception_type, error_msg):
    if logger.isEnabledFor(logging.ERROR):
        logger.error(error_msg, exc_info=True)

    raise exception_type(error_msg)



from pysqream.utils import get_ram_windows, get_ram_linux
import sys

__version__ = '3.2.1'
buf_maps, buf_views = [], []
WIN = True if sys.platform in ('win32', 'cygwin') else False
MAC = True if sys.platform in ('darwin') else False
PROTOCOL_VERSION = 8
SUPPORTED_PROTOCOLS = 6, 7, 8
BUFFER_SIZE = 100 * int(1e6)  # For setting auto-flushing on netrwork insert
ROWS_PER_FLUSH = 100000
DEFAULT_CHUNKSIZE = 0  # Dummy variable for some jsons
FETCH_MANY_DEFAULT = 1  # default parameter for fetchmany()
VARCHAR_ENCODING = 'ascii'

CYTHON = False # Cython IS NOT SUPPORTED
clean_sqream_errors = False
support_pandas = False
dbg = False

if WIN:
    get_ram = get_ram_windows()
elif MAC:
    get_ram = None
else:
    get_ram = get_ram_linux()

try:
    import pyarrow as pa
    from pyarrow import csv
    import numpy as np
    ARROW = True
except:
    ARROW = False
else:
    sqream_to_pa = {
        'ftBool':     pa.bool_(),
        'ftUByte':    pa.uint8(),
        'ftShort':    pa.int16(),
        'ftInt':      pa.int32(),
        'ftLong':     pa.int64(),
        'ftFloat':    pa.float32(),
        'ftDouble':   pa.float64(),
        'ftDate':     pa.timestamp('ns'),
        'ftDateTime': pa.timestamp('ns'),
        'ftVarchar':  pa.string(),
        'ftBlob':     pa.utf8(),
        'ftNumeric':  pa.decimal128(38, 11)
    }

# For encoding data to be sent to SQream using struct.pack() and for type checking by _set_val()
type_to_letter = {
    'ftBool': '?',
    'ftUByte': 'B',
    'ftShort': 'h',
    'ftInt': 'i',
    'ftLong': 'q',
    'ftFloat': 'f',
    'ftDouble': 'd',
    'ftDate': 'i',
    'ftDateTime': 'q',
    'ftVarchar': 's',
    'ftBlob': 's',
    'ftNumeric': '4i'
}

typecodes = {
    'ftBool': 'NUMBER',
    'ftUByte': 'NUMBER',
    'ftInt': 'NUMBER',
    'ftShort': 'NUMBER',
    'ftLong': 'NUMBER',
    'ftDouble': 'NUMBER',
    'ftFloat': 'NUMBER',
    'ftDate': 'DATETIME',
    'ftDateTime': 'DATETIME',
    'ftVarchar': 'STRING',
    'ftBlob': 'STRING',
    'ftNumeric': 'NUMBER'
}





def pytest_addoption(parser):
    parser.addoption("--ip", action="store", help="SQream Server ip", default="192.168.0.35")


def pytest_generate_tests(metafunc):
    metafunc.config.getoption("ip")



import logging
import sys

initiated = False


def Logger():
    return __init_logger()


def __init_logger():
    global initiated

    logger = logging.getLogger()
    if not initiated:
        logger.setLevel(logging.DEBUG)
        fh = logging.FileHandler("test_log")
        fh.setLevel(logging.DEBUG)
        ch = logging.StreamHandler(stream=sys.stdout)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        fh.setFormatter(formatter)
        ch.setFormatter(formatter)
        logger.addHandler(fh)
        logger.addHandler(ch)
        initiated = True
    return logger



import time
from datetime import datetime, date, timezone
from numpy.random import randint, uniform
from math import floor
from queue import Queue
from subprocess import Popen
from time import sleep
import threading, sys, os
import pytest
import pysqream
sys.path.append(os.path.abspath(__file__).rsplit('tests/', 1)[0] + '/tests/')
from base import TestBase, TestBaseWithoutBeforeAfter, Logger, connect_dbapi


q = Queue()
varchar_length = 10
nvarchar_length = 10
max_bigint = sys.maxsize if sys.platform not in ('win32', 'cygwin') else 2147483647


def generate_varchar(length):
    return ''.join(chr(num) for num in randint(32, 128, length))


col_types = {'bool', 'tinyint', 'smallint', 'int', 'bigint', 'real', 'double', 'date', 'datetime',
             'varchar({})'.format(varchar_length), 'nvarchar({})'.format(varchar_length)}


pos_test_vals = {'bool': (0, 1, True, False, 2, 3.6, 'test', (1997, 5, 9), (1997, 12, 12, 10, 10, 10)),
                 'tinyint': (randint(0, 255), randint(0, 255), 0, 255, True, False),
                 'smallint': (randint(-32768, 32767), 0, -32768, 32767, True, False),
                 'int': (randint(-2147483648, 2147483647), 0, -2147483648, 2147483647, True, False),
                 'bigint': (randint(1-max_bigint, max_bigint), 0, 1-max_bigint, max_bigint, True, False),
                 'real': (float('inf'), float('-inf'), float('+0'), float('-0'), round(uniform(1e-6, 1e6), 5), 837326.52428, True, False),   # float('nan')
                 'double': (float('inf'), float('-inf'), float('+0'), float('-0'), uniform(1e-6, 1e6), True, False),  # float('nan')
                 'date': (date(1998, 9, 24), date(2020, 12, 1), date(1997, 5, 9), date(1993, 7, 13)),
                 'datetime': (datetime(1001, 1, 1, 10, 10, 10), datetime(1997, 11, 30, 10, 10, 10), datetime(1987, 7, 27, 20, 15, 45), datetime(1993, 12, 20, 17, 25, 46)),
                 'varchar': (generate_varchar(varchar_length), generate_varchar(varchar_length), generate_varchar(varchar_length), 'b   '),
                 'nvarchar': ('א', 'א  ', '', 'ab א')}

neg_test_vals = {'tinyint': (258, 3.6, 'test',  (1997, 5, 9), (1997, 12, 12, 10, 10, 10)),
                 'smallint': (40000, 3.6, 'test', (1997, 5, 9), (1997, 12, 12, 10, 10, 10)),
                 'int': (9999999999, 3.6, 'test',  (1997, 5, 9), (1997, 12, 12, 10, 10, 10)),
                 'bigint': (92233720368547758070, 3.6, 'test', (1997, 12, 12, 10, 10, 10)),
                 'real': ('test', (1997, 12, 12, 10, 10, 10)),
                 'double': ('test', (1997, 12, 12, 10, 10, 10)),
                 'date': (5, 3.6, (-8, 9, 1), (2012, 15, 6), (2012, 9, 45), 'test', False, True),
                 'datetime': (5, 3.6, (-8, 9, 1, 0, 0, 0), (2012, 15, 6, 0, 0, 0), (2012, 9, 45, 0, 0, 0), (2012, 9, 14, 26, 0, 0), (2012, 9, 14, 13, 89, 0), 'test', False, True),
                 'varchar': (5, 3.6, (1, 2), (1997, 12, 12, 10, 10, 10), False, True),
                 'nvarchar': (5, 3.6, (1, 2), (1997, 12, 12, 10, 10, 10), False, True)}


class TestConnection(TestBaseWithoutBeforeAfter):

    def test_connection(self):

        Logger().info("Connection tests - wrong ip")
        try:
            pysqream.connect('123.4.5.6', 5000, 'master', 'sqream', 'sqream', False, False)
        except Exception as e:
            if "perhaps wrong IP?" not in repr(e):
                raise Exception("bad error message")

        # Logger().info("Connection tests - wrong port")
        # try:
        #     pysqream.connect(ip, 6000, 'master', 'sqream', 'sqream', False, False)
        # except Exception as e:
        #     if "Connection refused" not in repr(e):
        #         raise Exception("bad error message")

        Logger().info("Connection tests - wrong database")
        try:
            pysqream.connect(self.ip, 5000, 'wrong_db', 'sqream', 'sqream', False, False)
        except Exception as e:
            if "Database 'wrong_db' does not exist" not in repr(e):
                raise Exception("bad error message")

        Logger().info("Connection tests - wrong username")
        try:
            pysqream.connect(self.ip, 5000, 'master', 'wrong_username', 'sqream', False, False)
        except Exception as e:
            if "role 'wrong_username' doesn't exist" not in repr(e):
                raise Exception("bad error message")

        Logger().info("Connection tests - wrong password")
        try:
            pysqream.connect(self.ip, 5000, 'master', 'sqream', 'wrong_pw', False, False)
        except Exception as e:
            if "wrong password for role 'sqream'" not in repr(e):
                raise Exception("bad error message")

        Logger().info("Connection tests - close() function")
        con = connect_dbapi(self.ip)
        cur = con.cursor()
        con.close()
        try:
            cur.execute('select 1')
        except Exception as e:
          if "Connection has been closed" not in repr(e):
              raise Exception("bad error message")
        finally:
            cur.close()

        Logger().info("Connection tests - close_connection() function")
        con = connect_dbapi(self.ip)
        cur = con.cursor()
        con.close()
        try:
            cur.execute('select 1')
        except Exception as e:
            if "Connection has been closed" not in repr(e):
                raise Exception("bad error message")
        finally:
            cur.close()

        Logger().info("Connection tests - Trying to close a connection that is already closed with close()")
        con = connect_dbapi(self.ip)
        con.close()
        try:
            con.close()
        except Exception as e:
            if "Trying to close a connection that's already closed" not in repr(e):
                raise Exception("bad error message")
        #
        Logger().info("Connection tests - Trying to close a connection that is already closed with close_connection()")
        con = connect_dbapi(self.ip)
        con.close()
        try:
            con.close()
        except Exception as e:
            if "Trying to close a connection that's already closed" not in repr(e):
                raise Exception("bad error message")
        #
        # Logger().info("Connection tests - negative test for use_ssl=True")
        # try:
        #     pysqream.connect(self.ip, 5000, 'master', 'sqream', 'sqream', False, True)
        # except Exception as e:
        #     if "Using use_ssl=True but connected to non ssl sqreamd port" not in repr(e):
        #         raise Exception("bad error message")

        # Logger().info("Connection tests - positive test for use_ssl=True")
        # con = connect_dbapi(self.ip, False, True)
        # cur = con.cursor()
        # res = cur.execute('select 1').fetchall()[0][0]
        # if res != 1:
        #     if f'expected to get 1, instead got {res}' not in repr(e):
        #         raise Exception("bad error message")

        Logger().info("Connection tests - negative test for clustered=True")
        try:
            pysqream.connect(self.ip, 5000, 'master', 'sqream', 'sqream', True, False)
        except Exception as e:
            if "Connected with clustered=True, but apparently not a server picker port" not in repr(e):
                raise Exception("bad error message")

        Logger().info("Connection tests - positive test for clustered=True")
        con = pysqream.connect(self.ip, 3108, "master", "sqream", "sqream", clustered=True)
        cur = con.cursor()
        cur.execute('select 1')
        res = cur.fetchall()[0][0]
        if res != 1:
            if f'expected to get 1, instead got {res}' not in repr(e):
                raise Exception("bad error message")
        con.close()

        Logger().info("Connection tests - both clustered and use_ssl flags on True")
        con = connect_dbapi(self.ip, True, True)
        cur = con.cursor()
        res = cur.execute('select 1').fetchall()[0][0]
        if res != 1:
            if f'expected to get 1, instead got {res}' not in repr(e):
                raise Exception("bad error message")
        con.close()


class TestPositive(TestBase):

    def test_positive(self):

        cur = self.con.cursor()
        for col_type in col_types:
            trimmed_col_type = col_type.split('(')[0]

            Logger().info(f'Positive tests - Inserted values test for column type {col_type}')
            cur.execute(f"create or replace table test (t_{trimmed_col_type} {col_type})")
            for val in pos_test_vals[trimmed_col_type]:
                cur.execute('truncate table test')
                rows = [(val,)]
                cur.executemany("insert into test values (?)", rows)
                res = cur.execute("select * from test").fetchall()[0][0]

                # Compare
                if val != res:
                    if trimmed_col_type not in ('bool', 'varchar', 'date', 'datetime', 'real'):
                        Logger().info((repr(val), type(val), repr(res), type(res)))
                        raise Exception("TEST ERROR: No match between the expected result to the returned result. expected to get {}, instead got {} on datatype {}".format(repr(val), repr(res), trimmed_col_type))
                    elif trimmed_col_type == 'bool' and val != 0:
                        if res is not True:
                            raise Exception("TEST ERROR: No match between the expected result to the returned result. expected to get 'True', instead got {} on datatype {}".format(repr(res), trimmed_col_type))
                    elif trimmed_col_type == 'varchar' and val.strip() != res:
                        raise Exception("TEST ERROR: No match between the expected result to the returned result. expected to get {}, instead got {} on datatype {}".format(repr(val), repr(res), trimmed_col_type))
                    elif trimmed_col_type in ('date', 'datetime') and datetime(*val) != res and date(*val) != res:
                        raise Exception("TEST ERROR: No match between the expected result to the returned result. expected to get {}, instead got {} on datatype {}".format(repr(val), repr(res), trimmed_col_type))
                    elif trimmed_col_type == 'real' and abs(res-val) > 0.1:
                        # Single precision packing and unpacking is inaccurate:
                        # unpack('f', pack('f', 255759.83335))[0] == 255759.828125
                        raise Exception("TEST ERROR: No match between the expected result to the returned result. expected to get {}, instead got {} on datatype {}".format(repr(val), repr(res), trimmed_col_type))

            Logger().info(f'Positive tests - Null test for column type: {col_type}')
            cur.execute("create or replace table test (t_{} {})".format(trimmed_col_type, col_type))
            cur.executemany('insert into test values (?)', [(None,)])
            res = cur.execute('select * from test').fetchall()[0][0]
            if res not in (None,):
                raise Exception("TEST ERROR: Error setting null on column type: {}\nGot: {}, {}".format(trimmed_col_type, res, type(res)))

        Logger().info("Positive tests - Case statement with nulls")
        cur.execute("create or replace table test (xint int)")
        cur.executemany('insert into test values (?)', [(5,), (None,), (6,), (7,), (None,), (8,), (None,)])
        cur.executemany("select case when xint is null then 1 else 0 end from test")
        expected_list = [0, 1, 0, 0, 1, 0, 1]
        res_list = []
        res_list += [x[0] for x in cur.fetchall()]
        if expected_list != res_list:
            raise Exception("expected to get {}, instead got {}".format(expected_list, res_list))

        Logger().info("Positive tests - Testing select true/false")
        cur.execute("select false")
        res = cur.fetchall()[0][0]
        if res != 0:
            raise Exception("Expected to get result 0, instead got {}".format(res))
        cur.execute("select true")
        res = cur.fetchall()[0][0]
        if res != 1:
            raise Exception("Expected to get result 1, instead got {}".format(res))

        Logger().info("Positive tests - Running a statement when there is an open statement")
        cur.execute("select 1")
        sleep(10)
        res = cur.execute("select 1").fetchall()[0][0]
        if res != 1:
            raise Exception(f'expected to get result 1, instead got {res}')

        cur.close()


class TestNegative(TestBase):

    def test_negative(self):
        ''' Negative Set/Get tests '''

        cur = self.con.cursor()

        for col_type in col_types:
            if col_type == 'bool':
                continue
            Logger().info("Negative tests for column type: {}".format(col_type))
            trimmed_col_type = col_type.split('(')[0]
            Logger().info("prepare a table")
            cur.execute("create or replace table test (t_{} {})".format(trimmed_col_type, col_type))
            for val in neg_test_vals[trimmed_col_type]:
                Logger().info("Insert value {} into data type {}".format(repr(val), repr(trimmed_col_type)))
                rows = [(val,)]
                try:
                    cur.executemany("insert into test values (?)", rows)
                except Exception as e:
                    if "Error packing columns. Check that all types match the respective column types" not in repr(e):
                        raise Exception(f'bad error message')

        Logger().info("Negative tests - Inconsistent sizes test")
        cur.execute("create or replace table test (xint int, yint int)")
        try:
            cur.executemany('insert into test values (?, ?)', [(5,), (6, 9), (7, 8)])
        except Exception as e:
            if "Incosistent data sequences passed for inserting. Please use rows/columns of consistent length" not in repr(e):
                raise Exception(f'bad error message')

        Logger().info("Negative tests - Varchar - Conversion of a varchar to a smaller length")
        cur.execute("create or replace table test (test varchar(10))")
        try:
            cur.executemany("insert into test values ('aa12345678910')")
        except Exception as e:
            if "expected response statementPrepared but got" not in repr(e):
                            raise Exception(f'bad error message')

        Logger().info("Negative tests - Nvarchar - Conversion of a varchar to a smaller length")
        cur.execute("create or replace table test (test nvarchar(10))")
        try:
            cur.executemany("insert into test values ('aa12345678910')")
        except Exception as e:
            if "expected response executed but got" not in repr(e):
                raise Exception(f'bad error message')

        Logger().info("Negative tests - Incorrect usage of fetchmany - fetch without a statement")
        cur.execute("create or replace table test (xint int)")
        try:
            cur.fetchmany(2)
        except Exception as e:
            if "No open statement while attempting fetch operation" not in repr(e):
                raise Exception(f'bad error message')

        Logger().info("Negative tests - Incorrect usage of fetchall")
        cur.execute("create or replace table test (xint int)")
        cur.executemany("select * from test")
        try:
            cur.fetchall(5)
        except Exception as e:
            if "Bad argument to fetchall" not in repr(e):
                raise Exception(f'bad error message')

        Logger().info("Negative tests - Incorrect usage of fetchone")
        cur.execute("create or replace table test (xint int)")
        cur.executemany("select * from test")
        try:
            cur.fetchone(5)
        except Exception as e:
            if "Bad argument to fetchone" not in repr(e):
                raise Exception(f'bad error message')

        Logger().info("Negative tests - Multi statements test")
        try:
            cur.execute("select 1; select 1;")
        except Exception as e:
            if "expected one statement, got " not in repr(e):
                raise Exception(f'bad error message')

        Logger().info("Negative tests - Parametered query tests")
        params = 6
        cur.execute("create or replace table test (xint int)")
        cur.executemany('insert into test values (?)', [(5,), (6,), (7,)])
        try:
            cur.execute('select * from test where xint > ?', str(params))
        except Exception as e:
            if "Parametered queries not supported" not in repr(e):
                raise Exception(f'bad error message')

        Logger().info("Negative tests - running execute on a closed cursor")
        cur.close()
        cur = self.con.cursor()
        cur.close()
        try:
            cur.execute("select 1")
        except Exception as e:
            if "Cursor has been closed" not in repr(e):
                raise Exception(f'bad error message')


class TestFetch(TestBase):

    def test_fetch(self):
        cur = self.con.cursor()

        Logger().info("Fetch tests - positive fetch tests")
        cur.execute("create or replace table test (xint int)")
        cur.executemany('insert into test values (?)', [(1,), (2,), (3,), (4,), (5,), (6,), (7,), (8,), (9,), (10,)])
        # fetchmany(1) vs fetchone()
        cur.execute("select * from test")
        res = cur.fetchmany(1)[0][0]
        cur.execute("select * from test")
        res2 = cur.fetchone()[0]
        if res != res2:
            raise Exception(f"fetchmany(1) and fetchone() didn't return the same value. fetchmany(1) is {res} and fetchone() is {res2}")
        # fetchmany(-1) vs fetchall()
        cur.execute("select * from test")
        res3 = cur.fetchmany(-1)
        cur.execute("select * from test")
        res4 = cur.fetchall()
        if res3 != res4:
            raise Exception("fetchmany(-1) and fetchall() didn't return the same value. fetchmany(-1) is {} and fetchall() is {}".format(res3, res4))
        # fetchone() loop
        cur.execute("select * from test")
        for i in range(1, 11):
            x = cur.fetchone()[0]
            if x != i:
                raise Exception("fetchone() returned {} instead of {}".format(x, i))

        Logger().info("Fetch tests - combined fetch functions")
        cur.execute("create or replace table test (xint int)")
        cur.executemany('insert into test values (?)', [(1,), (2,), (3,), (4,), (5,), (6,), (7,), (8,), (9,), (10,)])
        cur.execute("select * from test")
        expected_list = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        res_list = []
        res_list.append(cur.fetchone()[0])
        res_list += [x[0] for x in cur.fetchmany(2)]
        res_list.append(cur.fetchone()[0])
        res_list += [x[0] for x in cur.fetchall()]
        if expected_list != res_list:
            raise Exception("expected to get {}, instead got {}".format(expected_list, res_list))

        Logger().info("Fetch tests - fetch functions after all the data has already been read")
        cur.execute("create or replace table test (xint int)")
        cur.executemany('insert into test values (?)', [(1,)])
        cur.execute("select * from test")
        x = cur.fetchone()[0]
        res = cur.fetchone()
        if res is not None:
            raise Exception(f"expected to get an empty result from fetchone, instead got {res}")
        res = cur.fetchall()
        if res != []:
            raise Exception(f"expected to get an empty result from fetchall, instead got {res}")
        res = cur.fetchmany(1)
        if res != []:
            raise Exception(f"expected to get an empty result from fetchmany, instead got {res}")

        cur.close()


class TestCursor(TestBaseWithoutBeforeAfter):

    def test_cursor(self):


        Logger().info("Cursor tests - running two statements on the same cursor connection")
        vals = [1]
        con = connect_dbapi(self.ip)
        cur = con.cursor()
        cur.execute("select 1")
        res1 = cur.fetchall()[0][0]
        vals.append(res1)
        cur.execute("select 1")
        res2 = cur.fetchall()[0][0]
        vals.append(res2)
        if not all(x == vals[0] for x in vals):
            raise Exception(f"expected to get result 1, instead got {res1} and {res2}")
        cur.close()

        Logger().info("Cursor tests - running a statement through cursor when there is an open statement")
        cur = con.cursor()
        cur.execute("select 1")
        sleep(10)
        cur.execute("select 1")
        res = cur.fetchall()[0][0]
        if res != 1:
            raise Exception(f"expected to get result 1, instead got {res}")
        cur.close()

        Logger().info("Cursor tests - fetch functions after all the data has already been read through cursor")
        cur = con.cursor()
        cur.execute("create or replace table test (xint int)")
        cur.executemany('insert into test values (?)', [(1,)])
        cur.execute("select * from test")
        x = cur.fetchone()[0]
        res = cur.fetchone()
        if res is not None:
            raise Exception("expected to get an empty result from fetchone, instead got {}".format(res))
        res = cur.fetchall()
        if res != []:
            raise Exception("expected to get an empty result from fetchall, instead got {}".format(res))
        res = cur.fetchmany(1)
        if res != []:
            raise Exception("expected to get an empty result from fetchmany, instead got {}".format(res))
        cur.close()

        Logger().info("Cursor tests - run a query through a cursor and close the connection directly")
        cur = con.cursor()
        cur.execute("select 1")
        con.close()
        if not cur.closed:
            raise Exception(f'Closed a connection after running a query through a cursor, but cursor is still open')


class TestString(TestBase):

    def test_string(self):

        cur = self.con.cursor()
        Logger().info("String tests - insert and return UTF-8")
        cur.execute("create or replace table test (xvarchar varchar(20))")
        cur.executemany('insert into test values (?)', [(u"hello world",), ("hello world",)])
        cur.execute("select * from test")
        res = cur.fetchall()
        if res[0][0] != res[1][0]:
            raise Exception("expected to get identical strings from select statement. instead got {} and {}".format(res[0][0], res[1][0]))

        Logger().info("String tests - strings with escaped characters")
        cur.execute("create or replace table test (xvarchar varchar(20))")
        values = [("\t",), ("\n",), ("\\n",), ("\\\n",), (" \\",), ("\\\\",), (" \nt",), ("'abd''ef'",), ("abd""ef",), ("abd\"ef",)]
        cur.executemany('insert into test values (?)', values)
        cur.executemany("select * from test")
        expected_list = ['', '', '\\n', '\\', ' \\', '\\\\', ' \nt', "'abd''ef'", 'abdef', 'abd"ef']
        res_list = []
        res_list += [x[0] for x in cur.fetchall()]
        if expected_list != res_list:
            raise Exception("expected to get {}, instead got {}".format(expected_list, res_list))

        cur.close()


class TestDatetime(TestBase):

    def test_datetime(self):

        cur = self.con.cursor()
        Logger().info("Datetime tests - insert different timezones datetime")
        t1 = datetime.strptime(datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M"), '%Y-%m-%d %H:%M')
        t2 = datetime.strptime(datetime.now().strftime("%Y-%m-%d %H:%M"), '%Y-%m-%d %H:%M')
        cur.execute("create or replace table test (xdatetime datetime)")
        cur.executemany('insert into test values (?)', [(t1,), (t2,)])
        cur.execute("select * from test")
        res = cur.fetchall()
        if res[0][0] == res[1][0]:
            raise Exception("expected to get different datetimes")

        Logger().info("Datetime tests - insert datetime with microseconds")
        t1 = datetime(1997, 5, 9, 4, 30, 10, 123456)
        t2 = datetime(1997, 5, 9, 4, 30, 10, 987654)
        cur.execute("create or replace table test (xdatetime datetime)")
        cur.executemany('insert into test values (?)', [(t1,), (t2,)])

        cur.close()


class TestThreads(TestBaseWithoutBeforeAfter):

    def connect_and_execute(self, num, con):
        cur = con.cursor()
        cur.execute("select {}".format(num))
        res = cur.fetchall()
        q.put(res)
        cur.close()

    def test_threads(self):
        con = connect_dbapi(self.ip)

        Logger().info("Thread tests - concurrent inserts with multiple threads through cursor")
        t1 = threading.Thread(target=self.connect_and_execute, args=(3, con,))
        t2 = threading.Thread(target=self.connect_and_execute, args=(3, con,))
        t1.start()
        t2.start()
        t1.join()
        t2.join()
        res1 = q.get()[0][0]
        res2 = q.get()[0][0]
        if res1 != res2:
            raise Exception("expected to get equal values. instead got res1 {} and res2 {}".format(res1, res2))

        con.close()


# def copy_tests():
#     global con
#     cur = con.cursor()
#     print("loading a csv file into a table through dbapi")
#     cur.execute("create or replace table t (xint1 int, xint2 int, xbigint1 bigint, xbigint2 bigint, xdouble1 double,"
#                 "xint3 int,xdouble2 double, xdate date, xdatetime datetime,xint4 int, xtext text, xint5 int, xint6 int)")
#     cur.csv_to_table(os.path.join(os.path.abspath("."), "t.csv"), "t", delimiter="|")
#     cur.execute("select count(*) from t")
#     res = cur.fetchall()[0][0]
#     if res != 2000:
#         raise Exception("expected to get 2000, instead got {}".format(res))



from datetime import datetime, date, timezone
from numpy.random import randint, uniform
from math import floor
from queue import Queue
from subprocess import Popen
from time import sleep
from decimal import Decimal, getcontext
import pytest
import threading, sys, os
import pysqream
sys.path.append(os.path.abspath(__file__).rsplit('tests/', 1)[0] + '/tests/')
from base import TestBase, TestBaseWithoutBeforeAfter, Logger, connect_dbapi


q = Queue()
varchar_length = 10
nvarchar_length = 10
precision = 38
scale = 10
max_bigint = sys.maxsize if sys.platform not in ('win32', 'cygwin') else 2147483647


def generate_varchar(length):
    return ''.join(chr(num) for num in randint(32, 128, length))


getcontext().prec = 38


col_types = ['bool', 'tinyint', 'smallint', 'int', 'bigint', 'real', 'double', 'date', 'datetime',
             'varchar({})'.format(varchar_length), 'nvarchar({})'.format(varchar_length),
             'numeric({},{})'.format(precision, scale)]

pos_test_vals = {'bool': (0, 1, True, False, 2, 3.6, 'test', (1997, 5, 9), (1997, 12, 12, 10, 10, 10)),
                 'tinyint': (randint(0, 255), randint(0, 255), 0, 255, True, False),
                 'smallint': (randint(-32768, 32767), 0, -32768, 32767, True, False),
                 'int': (randint(-2147483648, 2147483647), 0, -2147483648, 2147483647, True, False),
                 'bigint': (randint(1 - max_bigint, max_bigint), 0, 1 - max_bigint, max_bigint, True, False),
                 'real': (
                 float('inf'), float('-inf'), float('+0'), float('-0'), round(uniform(1e-6, 1e6), 5), 837326.52428,
                 True, False),  # float('nan')
                 'double': (float('inf'), float('-inf'), float('+0'), float('-0'), uniform(1e-6, 1e6), True, False),
                 # float('nan')
                 'date': (date(1998, 9, 24), date(2020, 12, 1), date(1997, 5, 9), date(1993, 7, 13), date(1001, 1, 1)),
                 'datetime': (datetime(1001, 1, 1, 10, 10, 10), datetime(1997, 11, 30, 10, 10, 10),
                              datetime(1987, 7, 27, 20, 15, 45), datetime(1993, 12, 20, 17, 25, 46)),
                 'varchar': (
                 generate_varchar(varchar_length), generate_varchar(varchar_length), generate_varchar(varchar_length),
                 'b   '),
                 'nvarchar': ('א', 'א  ', '', 'ab א'),
                 'numeric': (Decimal("0"), Decimal("1"), Decimal("1.1"), Decimal("-1"), Decimal("-1.0"),
                             Decimal("12345678901234567890.0123456789"))}

neg_test_vals = {'tinyint': (258, 3.6, 'test', (1997, 5, 9), (1997, 12, 12, 10, 10, 10)),
                 'smallint': (40000, 3.6, 'test', (1997, 5, 9), (1997, 12, 12, 10, 10, 10)),
                 'int': (9999999999, 3.6, 'test', (1997, 5, 9), (1997, 12, 12, 10, 10, 10)),
                 'bigint': (92233720368547758070, 3.6, 'test', (1997, 12, 12, 10, 10, 10)),
                 'real': ('test', (1997, 12, 12, 10, 10, 10)),
                 'double': ('test', (1997, 12, 12, 10, 10, 10)),
                 'date': (5, 3.6, (-8, 9, 1), (2012, 15, 6), (2012, 9, 45), 'test', False, True),
                 'datetime': (
                 5, 3.6, (-8, 9, 1, 0, 0, 0), (2012, 15, 6, 0, 0, 0), (2012, 9, 45, 0, 0, 0), (2012, 9, 14, 26, 0, 0),
                 (2012, 9, 14, 13, 89, 0), 'test', False, True),
                 'varchar': (5, 3.6, (1, 2), (1997, 12, 12, 10, 10, 10), False, True),
                 'nvarchar': (5, 3.6, (1, 2), (1997, 12, 12, 10, 10, 10), False, True),
                 'numeric': ('a')}


class TestPositive(TestBase):

    def test_positive(self):

        cur = self.con.cursor()
        Logger().info('positive tests')
        for col_type in col_types:
            trimmed_col_type = col_type.split('(')[0]

            Logger().info(f'Inserted values test for column type {col_type}')
            cur.execute(f"create or replace table test (t_{trimmed_col_type} {col_type})")
            for val in pos_test_vals[trimmed_col_type]:
                cur.execute('truncate table test')
                rows = [(val,)]
                cur.executemany("insert into test values (?)", rows)
                res = cur.execute("select * from test").fetchall()[0][0]
                # Compare
                error = False
                assert (
                        val == res or
                        (val != res and trimmed_col_type == 'bool' and val != 0 and res == True) or
                        (val != res and trimmed_col_type == 'varchar' and val != 0 and val.strip() == res) or
                        (val != res and trimmed_col_type == 'real' and val != 0 and abs(res - val) <= 0.1)
                )

            Logger().info(f'Null test for column type: {col_type}')
            cur.execute("create or replace table test (t_{} {})".format(trimmed_col_type, col_type))
            cur.executemany('insert into test values (?)', [(None,)])
            res = cur.execute('select * from test').fetchall()[0][0]
            assert res == None

        cur.close()

    def test_nulls(self):

        cur = self.con.cursor()
        Logger().info("Case statement with nulls")
        cur.execute("create or replace table test (xint int)")
        cur.executemany('insert into test values (?)', [(5,), (None,), (6,), (7,), (None,), (8,), (None,)])
        cur.executemany("select case when xint is null then 1 else 0 end from test")
        expected_list = [0, 1, 0, 0, 1, 0, 1]
        res_list = []
        res_list += [x[0] for x in cur.fetchall()]
        cur.close()
        assert expected_list == res_list

    def test_bool(self):

        cur = self.con.cursor()
        Logger().info("Testing select true/false")
        cur.execute("select false")
        res = cur.fetchall()[0][0]
        assert res == 0

        cur.execute("select true")
        res = cur.fetchall()[0][0]
        cur.close()
        assert res == 1

    def test_when_running(self):

        cur = self.con.cursor()
        Logger().info("Running a statement when there is an open statement")
        cur.execute("select 1")
        sleep(10)
        res = cur.execute("select 1").fetchall()[0][0]
        cur.close()
        assert res == 1


class TestNegative(TestBase):
    ''' Negative Set/Get tests '''

    def test_negative(self):

        cur = self.con.cursor()
        Logger().info('Negative tests')
        for col_type in col_types:
            if col_type == 'bool':
                continue
            trimmed_col_type = col_type.split('(')[0]
            cur.execute("create or replace table test (t_{} {})".format(trimmed_col_type, col_type))
            for val in neg_test_vals[trimmed_col_type]:
                rows = [(val,)]
                with pytest.raises(Exception) as e:
                    cur.executemany("insert into test values (?)", rows)
                assert "Error packing columns. Check that all types match the respective column types" in str(e.value)
        cur.close()

    def test_incosistent_sizes(self):

        cur = self.con.cursor()
        Logger().info("Inconsistent sizes test")
        cur.execute("create or replace table test (xint int, yint int)")
        with pytest.raises(Exception) as e:
            cur.executemany('insert into test values (?, ?)', [(5,), (6, 9), (7, 8)])
        cur.close()
        assert "Incosistent data sequences passed for inserting. Please use rows/columns of consistent length" in str(
            e.value)

    def test_varchar_conversion(self):

        cur = self.con.cursor()
        Logger().info("Varchar - Conversion of a varchar to a smaller length")
        cur.execute("create or replace table test (test varchar(10))")
        with pytest.raises(Exception) as e:
            cur.executemany("insert into test values ('aa12345678910')")
        cur.close()
        assert "expected response statementPrepared but got" in str(e.value)

    def test_nvarchar_conversion(self):

        cur = self.con.cursor()
        Logger().info("Nvarchar - Conversion of a varchar to a smaller length")
        cur.execute("create or replace table test (test nvarchar(10))")
        with pytest.raises(Exception) as e:
            cur.executemany("insert into test values ('aa12345678910')")
        cur.close()
        assert "expected response executed but got" in str(e.value)

    def test_incorrect_fetchmany(self):

        cur = self.con.cursor()
        Logger().info("Incorrect usage of fetchmany - fetch without a statement")
        cur.execute("create or replace table test (xint int)")
        with pytest.raises(Exception) as e:
            cur.fetchmany(2)
        cur.close()
        assert "No open statement while attempting fetch operation" in str(e.value)

    def test_incorrect_fetchall(self):

        cur = self.con.cursor()
        Logger().info("Incorrect usage of fetchall")
        cur.execute("create or replace table test (xint int)")
        cur.executemany("select * from test")
        with pytest.raises(Exception) as e:
            cur.fetchall(5)
        cur.close()
        assert "Bad argument to fetchall" in str(e.value)

    def test_incorrect_fetchone(self):

        cur = self.con.cursor()
        Logger().info("Incorrect usage of fetchone")
        cur.execute("create or replace table test (xint int)")
        cur.executemany("select * from test")
        with pytest.raises(Exception) as e:
            cur.fetchone(5)
        cur.close()
        assert "Bad argument to fetchone" in str(e.value)

    def test_multi_statement(self):

        cur = self.con.cursor()
        Logger().info("Multi statements test")
        with pytest.raises(Exception) as e:
            cur.execute("select 1; select 1;")
        cur.close()
        assert "expected one statement, got" in str(e.value)

    def test_parametered_query(self):

        cur = self.con.cursor()
        Logger().info("Parametered query tests")
        params = 6
        cur.execute("create or replace table test (xint int)")
        cur.executemany('insert into test values (?)', [(5,), (6,), (7,)])
        with pytest.raises(Exception) as e:
            cur.execute('select * from test where xint > ?', str(params))
        cur.close()
        assert "Parametered queries not supported" in str(e.value)

    def test_execute_closed_cursor(self):

        cur = self.con.cursor()
        Logger().info("running execute on a closed cursor")
        cur.close()
        try:
            cur.execute("select 1")
        except Exception as e:
            if "Cursor has been closed" not in repr(e):
                raise Exception(f'bad error message')


class TestFetch(TestBase):

    def test_fetch(self):

        cur = self.con.cursor()
        cur.execute("create or replace table test (xint int)")
        cur.executemany('insert into test values (?)', [(1,), (2,), (3,), (4,), (5,), (6,), (7,), (8,), (9,), (10,)])
        # fetchmany(1) vs fetchone()
        cur.execute("select * from test")
        res = cur.fetchmany(1)[0][0]
        cur.execute("select * from test")
        res2 = cur.fetchone()[0]
        cur.close()
        assert res == res2

        # fetchmany(-1) vs fetchall()
        cur = self.con.cursor()
        cur.execute("select * from test")
        res3 = cur.fetchmany(-1)
        cur.execute("select * from test")
        res4 = cur.fetchall()
        cur.close()
        assert res3 == res4

        # fetchone() loop
        cur = self.con.cursor()
        cur.execute("select * from test")
        for i in range(1, 11):
            x = cur.fetchone()[0]
            assert x == i
        cur.close()

    def test_combined_fetch(self):

        cur = self.con.cursor()
        cur.execute("create or replace table test (xint int)")
        cur.executemany('insert into test values (?)', [(1,), (2,), (3,), (4,), (5,), (6,), (7,), (8,), (9,), (10,)])
        cur.execute("select * from test")
        expected_list = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        res_list = []
        res_list.append(cur.fetchone()[0])
        res_list += [x[0] for x in cur.fetchmany(2)]
        res_list.append(cur.fetchone()[0])
        res_list += [x[0] for x in cur.fetchall()]
        cur.close()
        assert expected_list == res_list

    def test_fetch_after_data_read(self):

        cur = self.con.cursor()
        cur.execute("create or replace table test (xint int)")
        cur.executemany('insert into test values (?)', [(1,)])
        cur.execute("select * from test")
        x = cur.fetchone()[0]
        res = cur.fetchone()
        assert res is None

        res = cur.fetchall()
        assert res == []

        res = cur.fetchmany(1)
        assert res == []

        cur.close()


class TestCursor(TestBaseWithoutBeforeAfter):

    def test_cursor_through_clustered(self):
        con_clustered = pysqream.connect(self.ip, 3108, 'master', 'sqream', 'sqream', clustered=True)
        cur = con_clustered.cursor()
        assert cur.execute("select 1").fetchall()[0][0] == 1
        cur.close()

    def test_two_statements_same_cursor(self):
        vals = [1]
        con = connect_dbapi(self.ip)
        cur = con.cursor()
        cur.execute("select 1")
        res1 = cur.fetchall()[0][0]
        vals.append(res1)
        cur.execute("select 1")
        res2 = cur.fetchall()[0][0]
        vals.append(res2)
        cur.close()
        con.close()
        assert all(x == vals[0] for x in vals)

    def test_cursor_when_open_statement(self):
        con = connect_dbapi(self.ip)
        cur = con.cursor()
        cur.execute("select 1")
        sleep(10)
        cur.execute("select 1")
        res = cur.fetchall()[0][0]
        cur.close()
        con.close()
        assert res == 1

    def test_fetch_after_all_read(self):
        con = connect_dbapi(self.ip)
        cur = con.cursor()
        cur.execute("create or replace table test (xint int)")
        cur.executemany('insert into test values (?)', [(1,)])
        cur.execute("select * from test")
        x = cur.fetchone()[0]
        res = cur.fetchone()
        assert res is None

        res = cur.fetchall()
        assert res == []

        res = cur.fetchmany(1)
        assert res == []
        cur.close()
        con.close()


class TestString(TestBase):

    def test_insert_return_utf8(self):
        cur = self.con.cursor()
        cur.execute("create or replace table test (xvarchar varchar(20))")
        cur.executemany('insert into test values (?)', [(u"hello world",), ("hello world",)])
        cur.execute("select * from test")
        res = cur.fetchall()
        cur.close()
        assert res[0][0] == res[1][0]

    def test_strings_with_escaped_chars(self):
        cur = self.con.cursor()
        cur.execute("create or replace table test (xvarchar varchar(20))")
        values = [("\t",), ("\n",), ("\\n",), ("\\\n",), (" \\",), ("\\\\",), (" \nt",), ("'abd''ef'",), ("abd""ef",),
                  ("abd\"ef",)]
        cur.executemany('insert into test values (?)', values)
        cur.executemany("select * from test")
        expected_list = ['', '', '\\n', '\\', ' \\', '\\\\', ' \nt', "'abd''ef'", 'abdef', 'abd"ef']
        res_list = []
        res_list += [x[0] for x in cur.fetchall()]
        cur.close()
        assert expected_list == res_list


"""SQream Native Python API"""


from datetime import datetime, date, time as t
import time
from pysqream.globals import __version__
from pysqream.logger import log_and_raise, start_logging, stop_logging
from pysqream.connection import Connection


def enable_logs(log_path=None):
    start_logging(None if log_path is True else log_path)


def stop_logs():
    stop_logging()


def connect(host, port, database, username, password, clustered=False,
            use_ssl=False, service='sqream', log=False, reconnect_attempts=3, reconnect_interval=10):
    ''' Connect to SQream database '''
    if not isinstance(reconnect_attempts, int) or reconnect_attempts < 0:
        log_and_raise(Exception, f'reconnect attempts should be a positive integer, got : {reconnect_attempts}')
    if not isinstance(reconnect_interval, int) or reconnect_attempts < 0:
        log_and_raise(Exception, f'reconnect interval should be a positive integer, got : {reconnect_interval}')

    conn = Connection(host, port, clustered, use_ssl, log=log, base_connection=True,
                        reconnect_attempts=reconnect_attempts, reconnect_interval=reconnect_interval)
    conn.connect_database(database, username, password, service)

    return conn


## DBapi compatibility
#  -------------------
''' To fully comply to Python's DB-API 2.0 database standard. Ignore when using internally '''

# Type objects and constructors required by the DB-API 2.0 standard
Binary = memoryview
Date = date
Time = t
Timestamp = datetime


class _DBAPITypeObject:
    """DB-API type object which compares equal to all values passed to the constructor.
        https://www.python.org/dev/peps/pep-0249/#implementation-hints-for-module-authors
    """
    def __init__(self, *values):
        self.values = values

    def __eq__(self, other):
        return other in self.values


STRING = "STRING"
BINARY = _DBAPITypeObject("BYTES", "RECORD", "STRUCT")
NUMBER = _DBAPITypeObject("INTEGER", "INT64", "FLOAT", "FLOAT64", "NUMERIC",
                          "BOOLEAN", "BOOL")
DATETIME = _DBAPITypeObject("TIMESTAMP", "DATE", "TIME", "DATETIME")
ROWID = "ROWID"

typecodes = {
    'ftBool': 'NUMBER',
    'ftUByte': 'NUMBER',
    'ftInt': 'NUMBER',
    'ftShort': 'NUMBER',
    'ftLong': 'NUMBER',
    'ftDouble': 'NUMBER',
    'ftFloat': 'NUMBER',
    'ftDate': 'DATETIME',
    'ftDateTime': 'DATETIME',
    'ftVarchar': 'STRING',
    'ftBlob': 'STRING',
    'ftNumeric': 'NUMBER'
}


def DateFromTicks(ticks):
    return Date.fromtimestamp(ticks)


def TimeFromTicks(ticks):
    return Time(
        *time.localtime(ticks)[3:6]
    )  # localtime() returns a namedtuple, fields 3-5 are hr/min/sec


def TimestampFromTicks(ticks):
    return Timestamp.fromtimestamp(ticks)


# DB-API global parameters
apilevel = '2.0' 

threadsafety = 1 # Threads can share the module but not a connection

paramstyle = 'qmark'


# if __name__ == '__main__':
#
#     print('PySqream DB-API connector, version ', __version__)
#
#     conn = connect("192.168.0.35", 5000, "master", "sqream", "sqream")
    # cur = conn.cursor()
    # cur.execute("select 1")
    # res = cur.fetchall()
    # print(res)
    # conn.close()



from pysqream.logger import *
from pysqream.globals import PROTOCOL_VERSION, SUPPORTED_PROTOCOLS, clean_sqream_errors
import socket
import ssl
import sys
import array
from struct import pack, unpack
from threading import Lock


class SQSocket:
    ''' Extended socket class with some'''

    def __init__(self, ip, port, use_ssl=False):
        self.ip, self.port, self.use_ssl = ip, port, use_ssl
        self._setup_socket(ip, port)

    def _setup_socket(self, ip, port):

        self.s = socket.socket()
        if self.use_ssl:
            # Python 3.10 SSL fix
            # 3.10 has increased the default TLS security settings,
            # need to downgrade to be compatible with current versions of sqream
            if sys.version_info.minor >= 10:
                self.ssl_context = ssl._create_unverified_context()
                self.ssl_context.set_ciphers('DEFAULT')  # 'AES256-SHA', 'RSA'
                # self.ssl_context.verify_mode = ssl.VerifyMode.CERT_NONE
                # self.ssl_context.options &= ~ssl.OP_NO_SSLv3
                self.s = self.ssl_context.wrap_socket(self.s, server_hostname=ip)
            else:
                self.s = ssl.wrap_socket(self.s)
        try:
            self.timeout(10)
            self.s.connect((ip, port))
        except ConnectionRefusedError as e:
            log_and_raise(ConnectionRefusedError, "Connection refused, perhaps wrong IP?")
        except ConnectionResetError:
            log_and_raise(Exception, 'Trying to connect to an SSL port with use_ssl = False')
        except Exception as e:
            if 'timeout' in repr(e):
                log_and_raise(Exception, "Timeout when connecting to SQream, perhaps wrong IP?")
            elif '[SSL: UNKNOWN_PROTOCOL] unknown protocol' in repr(e) or '[SSL: WRONG_VERSION_NUMBER]' in repr(e):
                log_and_raise(Exception, 'Using use_ssl=True but connected to non ssl sqreamd port')
            elif 'EOF occurred in violation of protocol (_ssl.c:' in repr(e):
                log_and_raise(Exception, 'Using use_ssl=True but connected to non ssl sqreamd port')
            else:
                log_and_raise(Exception, e)
        else:
            self.timeout(None)

    # General socket / tls socket functionality
    #

    def _check_server_up(self, ip=None, port=None, use_ssl=None):

        try:
            SQSocket(ip or self.ip, port or self.port, use_ssl or self.use_ssl)
        except ConnectionRefusedError:
            log_and_raise(ConnectionRefusedError, f"Connection to SQream interrupted")

    def send(self, data):

        # print ("sending: ", data)
        # try:
        return self.s.send(data)

        # except BrokenPipeError:
        #    raise BrokenPipeError('No connection to SQream. Try reconnecting')

    def close(self):
        return self.s.close()

    def timeout(self, timeout='not passed'):

        if timeout == 'not passed':
            return self.s.gettimeout()

        self.s.settimeout(timeout)

    # Extended functionality
    #

    def reconnect(self, ip=None, port=None):

        self.s.close()
        self._setup_socket(ip or self.ip, port or self.port)


class Client:

    def __init__(self, socket):
        self.socket = socket

    def receive(self, byte_num, timeout=None):
        ''' Read a specific amount of bytes from a given socket '''

        data = bytearray(byte_num)
        view = memoryview(data)
        total = 0

        if timeout:
            self.socket.settimeout(timeout)

        while view:
            # Get whatever the socket gives and put it inside the bytearray
            received = self.socket.s.recv_into(view)
            if received == 0:
                log_and_raise(ConnectionRefusedError, f'SQreamd connection interrupted - 0 returned by socket')
            view = view[received:]
            total += received

        if timeout:
            self.socket.settimeout(None)

        return data

    def get_response(self, is_text_msg=True):
        ''' Get answer JSON string from SQream after sending a relevant message '''
        lock = Lock()

        # Getting 10-byte response header back
        with lock:
            header = self.receive(10)
        server_protocol = header[0]
        if server_protocol not in SUPPORTED_PROTOCOLS:
            log_and_raise(Exception,
                          f'Protocol mismatch, client version - {PROTOCOL_VERSION}, server version - {server_protocol}')
        # bytes_or_text =  header[1]
        message_len = unpack('q', header[2:10])[0]

        with lock:
            receive = self.receive(message_len).decode(
                'utf8') if is_text_msg else self.receive(message_len)

        return receive

    # Non socket aux. functionality
    #
    def generate_message_header(self, data_length, is_text_msg=True, protocol_version=PROTOCOL_VERSION):
        """Generate SQream's 10 byte header prepended to any message"""

        return pack('bb', protocol_version, 1 if is_text_msg else 2) + pack(
            'q', data_length)

    def validate_response(self, response, expected):

        if expected not in response:
            # Color first line of SQream error (before the haskell thingy starts) in Red
            response = '\033[31m' + (response.split('\\n')[0] if clean_sqream_errors else response) + '\033[0m'
            log_and_raise(Exception, f'\nexpected response {expected} but got:\n\n {response}')

    def send_string(self, json_cmd, get_response=True, is_text_msg=True, sock=None):
        ''' Encode a JSON string and send to SQream. Optionally get response '''

        # Generating the message header, and sending both over the socket
        printdbg(f'string sent: {json_cmd}')
        self.socket.send(self.generate_message_header(len(json_cmd)) + json_cmd.encode('utf8'))

        if get_response:
            return self.get_response(is_text_msg)



import time
import threading
from pysqream.SQSocket import Client


class PingLoop(threading.Thread):
    def __init__(self, conn):
        self.conn = conn
        self.client = Client(self.conn)
        super(PingLoop, self).__init__()
        self.done = False

    def run(self):
        json_cmd = '{"ping":"ping"}'
        binary = self.client.generate_message_header(len(json_cmd)) + json_cmd.encode('utf8')
        while self.sleep():
            conn = self.conn
            try:
                conn.s.send(binary)
            except:
                self.done = True

    def halt(self):
        self.done = True

    def sleep(self):
        if self.done:
            return False
        count = 0
        while (count < 100):
            count = count + 1
            time.sleep(.1)
            if self.done:
                return False
        return True


def _start_ping_loop(conn):
    ping_loop = PingLoop(conn)
    ping_loop.start()
    return ping_loop


def _end_ping_loop(ping_loop):
    if ping_loop is not None:
        ping_loop.halt()
        ping_loop.join()


import socket, ssl, selectors, json, datetime, sys, itertools as it
import logging, time, traceback
from struct import pack, unpack
from collections import deque


PROTOCOL_VERSION = 8
VARCHAR_ENCODING = 'ascii'


def printdbg(*debug_print, dbg = True):
    if dbg:
        print(*debug_print)


sql_to_sqream_type = {
    'bool'     : ('ftBool', 1),
    'tinyint'  : ('ftUByte', 1),
    'smallint' : ('ftShort', 2),
    'int'      : ('ftInt', 4),
    'bigint'   : ('ftLong', 8),
    'real'     : ('ftFloat', 4),
    'float'    : ('ftDouble', 8),   # Yeah I know
    'double'   : ('ftDouble', 8),
    'date'     : ('ftDate', 4),
    'datetime' : ('ftDateTime', 8),
    'varchar'  : ('ftVarchar', None),
    'text'     : ('ftBlob', 0),
}


class MockException(Exception):
    pass


class MockSock:
    ''' Mock server connection class '''
    
    def __init__(self, ip = '127.0.0.1', port = 5000, use_ssl=False):
        self.ip, self.port, self.use_ssl = ip, port, use_ssl
        # self.db = sqlite3.connect('memory')
        self.created_tables = {}
        self.table_meta = {}
        self.table_data = {}
        self.sel = selectors.DefaultSelector()
        self.ss = socket.socket()
        if self.use_ssl:
            self.ss = ssl.wrap_socket(self.s)

        self.ss.bind((ip, port))
        # self.ss.setblocking(False)
        self.ss.listen()
        self.accept()
        # self.sel.register(self.ss, selectors.EVENT_READ, accept)
        

    def accept(self, blocking = True):
        ''' Distributing connections to sockets '''

        print (f'Listening on {self.port}')
        conn, addr = self.ss.accept()
        print (f'Connection accepted on address: {addr}')
        # conn.setblocking(blocking)
        # sel.register(conn, selectors.EVENT_READ, read)
        self.start(conn)

        
    def start(self, conn):

        stmt_id = 1
        expecting = 'connectDatabase'
        while True:
            ''' 
            # Selector path
            for key, mask in sel.select():
                key.data(key.fileobj, mask)
            # '''

            header = conn.recv(10)
            if not header:
                continue
            '''
            client_protocol = header[0]
            if client_protocol != PROTOCOL_VERSION
                raise MockException(f'Protocol mismatch, client version - {client_protocol}, server version - {PROTOCOL_VERSION}')
            bytes_or_text =  header[1]
            # '''
            printdbg ("header: ", header)
            message_len = unpack('q', header[2:10])[0]
            printdbg(f'message len:{message_len}')
            data = conn.recv(message_len)
            self.msg = json.loads(data.decode('utf8'))
            printdbg(f'received: {self.msg}')  
            
            # Close Connection requested
            if 'closeConnection' in self.msg:
                rsp = f'{{"connectionClosed":"connectionClosed"}}'
                self.send_json(conn, rsp)
                break      

            # Not a request to close connection, fall into state machine
            self.validate_msg(expecting)

            # Connect database
            if 'connectDatabase' in self.msg:
                rsp = f'{{"connectionId":1,"databaseConnected":"databaseConnected","varcharEncoding":"{VARCHAR_ENCODING}"}}'
                self.send_json(conn, rsp)
                expecting = 'getStatementId'
                continue

            # Get statement ID
            if 'getStatementId' in self.msg:
                rsp = f'{{"statementId":{stmt_id}}}'
                self.send_json(conn, rsp)
                stmt_id +=1
                expecting = 'prepareStatement'
                continue

            # Prepare statement
            if 'prepareStatement' in self.msg:
                query = self.msg['prepareStatement'].lower()
                self.query_type = query.split()[0]

                if self.query_type == 'create':
                    self.generate_table_metadata(query)

                    '''
                    printdbg(f'\nMeta for create query:')
                    for meta in self.table_meta["bobo"]:
                        printdbg(meta)
                    '''

                elif self.query_type == 'insert':
                    table_name = query.split()[2].lower()
                    if table_name not in self.table_meta:
                        raise MockException(f"Trying to insert to table {table_name} which wasn't created")
                    self.table_data[table_name] = deque()

                elif self.query_type == 'select':
                    try:
                        table_name = query.split()[3].lower()
                    except IndexError:
                        # None table select, e.g. 'select 1'
                        self.query_type = 'select_const'
                    else:
                        self.query_type = 'select_table'

                else:
                    # Query didn't start with create, select or insert
                    raise MockException(f"Query started with {self.query_type}, u trollin'?")


                rsp = f'{{"ip":"127.0.0.1","listener_id":0,"port":5000,"port_ssl":5001,"reconnect":false,"statementPrepared":true}}'
                self.send_json(conn, rsp)
                expecting = 'execute'
                continue

            # Execute
            if 'execute' in self.msg:
                rsp = f'{{"executed":"executed"}}'
                self.send_json(conn, rsp)
                expecting = 'queryTypeIn'
                continue

            # Insert query metadata
            if 'queryTypeIn' in self.msg:
                if self.query_type in ('create', 'select_const', 'select_table'):
                    # Only giving meta for insert queries
                    rsp = f'{{"queryType":[]}}' 
                    expecting = 'queryTypeOut'
                else:
                    # Query type is insert
                    rsp = f'{{"queryType":[{self.table_meta[table_name][0]}]}}' 
                    # printdbg (f'querytypeIn reponse:\n{rsp}')
                    expecting = 'put', 'closeStatement'

                self.send_json(conn, rsp)
                continue

            # Select query metadata
            if 'queryTypeOut' in self.msg:
                if self.query_type in ('create', 'insert') or self.query_type == 'select_const':
                    # Not a select from table query
                    rsp = f'{{"queryTypeNamed":[]}}'
                    expecting = 'closeStatement'
                else:
                    # Select from table
                    rsp = f'{{"queryTypeNamed":[{self.table_meta[table_name][0]}]}}' 
                    expecting = 'fetch', 'closeStatement'

                self.send_json(conn, rsp)
                continue

            # Sending binary buffer via netowrk insert
            if 'put' in self.msg:
                rows_num      = int(self.msg['put'])
                                
                # Get binary data
                binary_header = conn.recv(10)
                binary_len    = unpack('q', binary_header[2:10])[0]
                binary_data   = self.receive(conn, binary_len)
                # binary_data   = conn.recv(unpack('q', conn.recv(10)[2:10])[0])

                # Calculating colSzs
                col_sizes     = self.calculate_col_szs(table_name, rows_num, binary_data) 

               # Verify size is equal the sum of col_sizes
                self.table_data[table_name].append((rows_num, col_sizes, binary_data))
                # printdbg (f'binary data appended: {self.table_data}')
                rsp = f'{{"putted":"putted"}}'  
                self.send_json(conn, rsp)
                expecting = 'put', 'closeStatement'
                continue

            # Fetching inserted data from table
            if 'fetch' in self.msg:
                rows, col_szs, buf = self.table_data[table_name].popleft() if self.table_data[table_name] else (0, '', b'')
                '{"colSzs":[],"rows":0}'
                rsp = f'{{"colSzs":[{col_szs}],"rows":{rows}}}'  
                self.send_json(conn, rsp)
                conn.send(self.generate_message_header(len(buf)) + buf)
                
                expecting = 'fetch', 'closeStatement'
                continue

            # Close statement
            if 'closeStatement' in self.msg:
                rsp = f'{{"statementClosed":"statementClosed"}}'  
                self.send_json(conn, rsp)
                expecting = 'getStatementId'
                continue


    def send_json(self, conn, response):

        printdbg(f'\033[94msending: {response}\033[0m' + ('\n' if 'statementClosed' in response else ''))  
        return conn.send(self.generate_message_header(len(response)) + response.encode('utf8'))
    

    def receive(self, sock, byte_num, timeout=None):
        ''' Read a specific amount of bytes from a given socket '''

        data = bytearray(byte_num)
        view = memoryview(data)
        total = 0

        if timeout:
            sock.settimeout(timeout)
        
        while view:
            # Get whatever the socket gives and put it inside the bytearray
            received = sock.recv_into(view)
            if received == 0:
                raise ConnectionRefusedError('Client connection interrupted - 0 returned by socket')
            view = view[received:]
            total += received

        if timeout:
            sock.settimeout(None)
            
        
        return data


    # Non socket aux. functionality
    #

    def generate_message_header(self, data_length, is_text_msg=True, protocol_version=PROTOCOL_VERSION):
        ''' Generate SQream's 10 byte header prepended to any message '''

        return pack('bb', protocol_version, 1 if is_text_msg else 2) + pack('q', data_length)


    def validate_msg(self, expected, msg = None):

        expected = (expected,) if isinstance(expected, str) else expected
        if not any(expected_msg in (msg or self.msg) for expected_msg in expected):
            raise MockException(f'\nExpected {expected} json but got:\n{msg}')

            
    # Query parsing
    #  

    # Create
    def generate_table_metadata(self, query):
        ''' SQLite does not supply metadata after select query, keeping it during create '''

        table_name            = query[query.index('table') + 6 : query.index('(')].strip()
        names, types, is_null = zip(*((item.split()[0], item.split()[1], item.split()[-2]) for item in query[query.index('(')+1:-1].split(',')))
        types                 = [col_type.split('(') for col_type in types]
        
        self.cols_meta = [
            ('true' if col_type[0] == 'text' else 'false', 
             names[idx], 
             'false' if is_null[idx] == 'not' else 'true',
             sql_to_sqream_type[col_type[0]][0], 
             sql_to_sqream_type[col_type[0]][1] if col_type[0] != 'varchar' else int(col_type[1][:-1]) 
            )
            for idx, col_type in enumerate(types)
         ]

        # Col sizes template, used for the response to "fetch"
        # text columns sizes need to be extracted deliberately from the binary buffer
        col_szs_tups = ((1*(meta[2] == 'true'), 4*(meta[0] == 'true'), meta[-1] if meta[-1] !=0 else -1) for meta in self.cols_meta)
        col_szs      = [num for tup in col_szs_tups for num in tup if num !=0]

        # {"isTrueVarChar":{is_tvc},"name":"{name}","nullable":{is_nul},"type":["{col_type}",{col_size},0]}
        col_meta_st = '{{"isTrueVarChar":{},"name":"{}","nullable":{},"type":["{}",{},0]}}'
        
        self.table_meta[table_name] = [','.join(col_meta_st.format(*col_meta) for col_meta in self.cols_meta), col_szs]
        

    # Insert
    def calculate_col_szs(self, table_name, rows_num, binary_data):
        '''  The size of most columns is known in advance, except text columns '''

        col_sizes =  [size * rows_num for size in self.table_meta[table_name][1]]
        text_sizes_start = 0
        for idx in range(len(col_sizes)-1):
            if col_sizes[idx+1] < 0:
                # Next one up is an unknown size of a text data column, here cometh thy money
                col_sizes[idx+1] = sum(unpack(f'{rows_num}i', binary_data[text_sizes_start : text_sizes_start + col_sizes[idx]]))
                # printdbg(f'text col size: {col_sizes[idx+1]}')

            text_sizes_start += col_sizes[idx]


        return ','.join(str(size) for size in col_sizes)

    # Select



class MockListener:
    ''' Accept connections and create MockSock classes to handle them '''

    def __init__(self, ip = '127.0.0.1', port = 5000, use_ssl=False):
        self.ip, self.port, self.use_ssl = ip, port, use_ssl
        # self.db = sqlite3.connect('memory')
        self.sel = selectors.DefaultSelector()
        self.ss = socket.socket()
        if self.use_ssl:
            self.ss = ssl.wrap_socket(self.s)

        self.ss.bind((ip, port))
        # self.ss.setblocking(False)
        self.ss.listen()
        

    def accept(self):

        print (f'Listening on {self.port}')
        conn, addr = self.ss.accept()
        print (f'Connection accepted on address: {addr}')
        # conn.setblocking(blocking)
        # sel.register(conn, selectors.EVENT_READ, read)
        self.start(conn)



def bind(ip = '127.0.0.1', port = 5000):

    mock_sock = MockSock(ip, port)

    return mock_sock


if __name__ == '__main__':

    args = sys.argv
    # ip, port = args[1:]
    ip = '127.0.0.1'
    # ip = ''
    mock = bind(ip, 5000)



from mmap import mmap, ACCESS_READ 
from sys import exit


def is_covered():
    '''
    Checking a commit coverage report generated by diff-cover, and parsing
    to see if any lines in the commit are not ocvered. Exiting with -1 if uncovered lines
    are found
    '''

    cov_xml = open('commit_coverage.html', 'r+') 
    cov_map = mmap(cov_xml.fileno(), 0)

    missing = b'Missing</b>: ' 
    missing_loc = cov_map.find(missing)

    if missing_loc != -1:
        cov_map.seek(missing_loc + len(missing)) 

        if int(cov_map.read(2)) > 0:
            print ('Some lines not covered in commit')
            exit(-1)



if __name__ == '__main__':

    is_covered()



import re
from packaging import version
from subprocess import Popen, PIPE


def get_ram_linux():
    vmstat, err = Popen('vmstat -s'.split(), stdout=PIPE, stderr=PIPE).communicate()

    return int(vmstat.splitlines()[0].split()[0])


def get_ram_windows():
    pass


## Version compare
def version_compare(v1, v2) :
    if (v2 is None or v1 is None):
        return None
    r1 = re.search("\\d{4}(\\.\\d+)+", v1)
    r2 = re.search("\\d{4}(\\.\\d+)+", v2)
    if (r2 is None or r1 is None):
        return None
    v1 = version.parse(r1.group(0))
    v2 = version.parse(r2.group(0))
    return -1 if v1 < v2 else 1 if v1 > v2 else 0


class Error(Exception):
    pass

class Warning(Exception):
    pass

class InterfaceError(Error):
    pass


class DatabaseError(Error):
    pass


class DataError(DatabaseError):
    pass


class OperationalError(DatabaseError):
    pass


class IntegrityError(DatabaseError):
    pass


class InternalError(DatabaseError):
    pass


class ProgrammingError(DatabaseError):
    pass


class NotSupportedError(DatabaseError):
    pass


import pytest
import socket
import sys
import os
import pysqream
from pytest_logger import Logger


def connect_dbapi(ip, clustered=False, use_ssl=False):
    port = (3109 if use_ssl else 3108) if clustered else (5001 if use_ssl else 5000)
    return pysqream.connect(ip, port, 'master', 'sqream', 'sqream', clustered, use_ssl)


class TestBase():

    @pytest.fixture()
    def ip(self, pytestconfig):
        return pytestconfig.getoption("ip")

    @pytest.fixture(autouse=True)
    def Test_setup_teardown(self, ip):
        ip = ip if ip else socket.gethostbyname(socket.gethostname())
        Logger().info("Before Scenario")
        Logger().info(f"Connect to server {ip}")
        self.con = connect_dbapi(ip)
        yield
        Logger().info("After Scenario")
        self.con.close()
        Logger().info(f"Close Session to server {ip}")


class TestBaseWithoutBeforeAfter():
    @pytest.fixture()
    def ip(self, pytestconfig):
        return pytestconfig.getoption("ip")

    @pytest.fixture(autouse=True)
    def Test_setup_teardown(self, ip):
        self.ip = ip if ip else socket.gethostbyname(socket.gethostname())
        yield


from pysqream.utils import version_compare
import json
from pysqream.globals import BUFFER_SIZE, ROWS_PER_FLUSH, DEFAULT_CHUNKSIZE, FETCH_MANY_DEFAULT, typecodes, type_to_letter, \
    ARROW, pa, csv
from pysqream.column_buffer import ColumnBuffer
from pysqream.ping import PingLoop, _start_ping_loop, _end_ping_loop
from pysqream.logger import *
from pysqream.utils import NotSupportedError, ProgrammingError, InternalError, IntegrityError, OperationalError, DataError, \
    DatabaseError, InterfaceError, Warning, Error
from pysqream.casting import lengths_to_pairs, sq_date_to_py_date, sq_datetime_to_py_datetime, sq_numeric_to_decimal
from pysqream.SQSocket import Client
import time


class Cursor:

    def __init__(self, conn, cursors):

        self.conn = conn
        self.cursors = cursors
        self.s = self.conn.s
        self.client = Client(self.s)
        self.version = self.conn.version
        self.open_statement = False
        self.closed = False
        self.buffer = ColumnBuffer(BUFFER_SIZE)  # flushing buffer every BUFFER_SIZE bytes
        self.ping_loop = None
        self.stmt_id = None  # For error handling when called out of order
        self.statement_type = None
        self.arraysize = FETCH_MANY_DEFAULT
        self.rowcount = -1  # DB-API property
        self.more_to_fetch = False
        self.parsed_rows = []
        self.row_size = 0
        self.rows_per_flush = 0
        self.lastrowid = None
    
    def get_statement_type(self):

        return self.statement_type

    def get_statement_id(self):

        return self.stmt_id

    def _execute_sqream_statement(self, stmt):
        self.latest_stmt = stmt

        if self.open_statement:
            self.close_stmt()
        self.open_statement = True

        self.more_to_fetch = True

        self.stmt_id = json.loads(self.client.send_string('{"getStatementId" : "getStatementId"}'))["statementId"]
        comp = version_compare(self.version, "2020.3.1")
        if (comp is not None and comp > -1):
            self.ping_loop = _start_ping_loop(self.conn)
        stmt_json = json.dumps({"prepareStatement": stmt, "chunkSize": DEFAULT_CHUNKSIZE})
        res = self.client.send_string(stmt_json)

        self.client.validate_response(res, "statementPrepared")
        self.lb_params = json.loads(res)
        if self.lb_params.get('reconnect'):  # Reconnect exists and issued, otherwise False / None

            # Close socket, open a new socket with new port/ip sent be the reconnect response
            self.s.reconnect(
                self.lb_params['ip'], self.lb_params['port_ssl']
                if self.conn.use_ssl else self.lb_params['port'])

            # Send reconnect and reconstruct messages
            reconnect_str = '{{"service": "{}", "reconnectDatabase":"{}", "connectionId":{}, "listenerId":{},"username":"{}", "password":"{}"}}'.format(
                self.conn.service, self.conn.database, self.conn.connection_id,
                self.lb_params['listener_id'], self.conn.username, self.conn.password)
            self.client.send_string(reconnect_str)
            self.client.send_string('{{"reconstructStatement": {}}}'.format(
                self.stmt_id))

        # Reconnected/reconstructed if needed,  send  execute command
        self.client.validate_response(self.client.send_string('{"execute" : "execute"}'), 'executed')

        # Send queryType message/s
        res = json.loads(self.client.send_string('{"queryTypeIn": "queryTypeIn"}'))
        self.column_list = res.get('queryType', '')

        if not self.column_list:
            res = json.loads(
                self.client.send_string('{"queryTypeOut" : "queryTypeOut"}'))
            self.column_list = res.get('queryTypeNamed', '')
            if not self.column_list:
                self.statement_type = 'DML'
                self.close_stmt()
                return

            self.statement_type = 'SELECT' if self.column_list else 'DML'
            self.result_rows = []
            self.unparsed_row_amount = 0
            self.data_columns = []
        else:
            self.statement_type = 'INSERT'

        # {"isTrueVarChar":false,"nullable":true,"type":["ftInt",4,0]}
        self.col_names, self.col_tvc, self.col_nul, self.col_type_tups = \
            list(zip(*[(col.get("name", ""), col["isTrueVarChar"], col["nullable"], col["type"]) for col in self.column_list]))
        self.col_names_map = {
            name: idx
            for idx, name in enumerate(self.col_names)
        }

        if self.statement_type == 'INSERT':
            self.col_types = [type_tup[0] for type_tup in self.col_type_tups]
            self.col_sizes = [type_tup[1] for type_tup in self.col_type_tups]
            self.col_scales = [type_tup[2] for type_tup in self.col_type_tups]
            self.row_size = sum(self.col_sizes) + sum(
                self.col_nul) + 4 * sum(self.col_tvc)
            self.rows_per_flush = ROWS_PER_FLUSH
            self.buffer.init_buffers(self.col_sizes, self.col_nul)

        # if self.statement_type == 'SELECT':
        self.parsed_rows = []
        self.parsed_row_amount = 0

        if logger.isEnabledFor(logging.INFO):
            logger.info \
                (f'Executing statement over connection {self.conn.connection_id} with statement id {self.stmt_id}:\n{stmt}')

    def _fill_description(self):
        """Getting parameters for the cursor's 'description' attribute, even for
           a query that returns no rows. For each column, this includes:
           (name, type_code, display_size, internal_size, precision, scale, null_ok)"""

        if self.statement_type != 'SELECT':
            self.description = None
            return self.description

        self.description = []
        for col_name, col_nullalbe, col_type_tup in zip(
                self.col_names, self.col_nul, self.col_type_tups):
            type_code = typecodes[
                col_type_tup[0]]  # Convert SQream type to DBAPI identifier
            display_size = internal_size = col_type_tup[
                1]  # Check if other size is available from API
            precision = 38
            scale = col_type_tup[2]

            self.description.append(
                (col_name, type_code, display_size, internal_size, precision,
                 scale, col_nullalbe))

        return self.description

    def _send_columns(self, cols=None, capacity=None):
        """Perform network insert - "put" json, header, binarized columns. Used by executemany()"""

        cols = cols or self.cols
        cols = cols if isinstance(cols, (list, tuple, set, dict)) else list(cols)

        capacity = capacity or self.capacity

        # Send columns and metadata to be packed into our buffer
        packed_cols = self.buffer.pack_columns(cols, capacity, self.col_types,
                                               self.col_sizes, self.col_nul,
                                               self.col_tvc, self.col_scales)
        del cols
        byte_count = sum(len(packed_col) for packed_col in packed_cols)

        # Sending put message and binary header
        self.client.send_string(f'{{"put":{capacity}}}', False)
        self.s.send((self.client.generate_message_header(byte_count, False)))

        # Sending packed data (binary buffer)
        for packed_col in packed_cols:
            self.s.send((packed_col))

        self.client.validate_response(self.client.get_response(), '{"putted":"putted"}')
        del packed_cols

    def _fetch(self, sock=None):

        sock = sock or self.s
        # JSON correspondence
        res = self.client.send_string('{"fetch" : "fetch"}')
        self.client.validate_response(res, "colSzs")
        fetch_meta = json.loads(res)
        num_rows_fetched, column_sizes = fetch_meta['rows'], fetch_meta['colSzs']
        if num_rows_fetched == 0:
            self.close_stmt()
            return num_rows_fetched

        # Get preceding header
        self.client.receive(10)

        # Get data as memoryviews of bytearrays.
        unsorted_data_columns = [memoryview(self.client.receive(size)) for idx, size in enumerate(column_sizes)]

        # Sort by columns, taking a memoryview and casting to the proper type
        self.data_columns = []

        for type_tup, nullable, tvc in zip(self.col_type_tups, self.col_nul,
                                           self.col_tvc):
            column = []
            if nullable:
                column.append(unsorted_data_columns.pop(0))
            if tvc:
                column.append(unsorted_data_columns.pop(0).cast('i'))

            column.append(unsorted_data_columns.pop(0))
            if type_tup[0] not in ('ftVarchar', 'ftBlob', 'ftNumeric'):
                column[-1] = column[-1].cast(type_to_letter[type_tup[0]])
            else:
                column[-1] = column[-1].tobytes()
            self.data_columns.append(column)

        self.unparsed_row_amount = num_rows_fetched

        return num_rows_fetched

    def _parse_fetched_cols(self, queue=None):
        """Used by _fetch_and_parse"""

        self.extracted_cols = []

        if not self.data_columns:
            return self.extracted_cols

        for idx, raw_col_data in enumerate(self.data_columns):
            # Extract data according to column type
            if self.col_tvc[idx]:  # nvarchar
                nvarc_sizes = raw_col_data[1 if self.col_nul[idx] else 0]
                col = [
                    raw_col_data[-1][start:end].decode('utf8')
                    for (start, end) in lengths_to_pairs(nvarc_sizes)
                ]
            elif self.col_type_tups[idx][0] == "ftVarchar":
                varchar_size = self.col_type_tups[idx][1]
                col = [
                    raw_col_data[-1][idx:idx + varchar_size].decode(
                        self.conn.varchar_enc, "ignore").replace('\x00', '').rstrip()
                    for idx in range(0, len(raw_col_data[-1]), varchar_size)
                ]
            elif self.col_type_tups[idx][0] == "ftDate":
                col = [sq_date_to_py_date(d) for d in raw_col_data[-1]]
            elif self.col_type_tups[idx][0] == "ftDateTime":
                col = [sq_datetime_to_py_datetime(d) for d in raw_col_data[-1]]
            elif self.col_type_tups[idx][0] == "ftNumeric":
                scale = self.col_type_tups[idx][2]
                col = [
                    # sq_numeric_to_decimal(bytes_to_bigint(raw_col_data[-1][idx:idx + 16]), scale)
                    sq_numeric_to_decimal(raw_col_data[-1][idx:idx + 16], scale)
                    for idx in range(0, len(raw_col_data[-1]), 16)
                ]

            else:
                col = raw_col_data[-1]

            # Fill Nones if / where needed
            if self.col_nul[idx]:
                nulls = raw_col_data[0]  # .tolist()
                col = [
                    item if not null else None
                    for item, null in zip(col, nulls)
                ]
            else:
                pass

            self.extracted_cols.append(col)

        # Done with the raw data buffers
        self.unparsed_row_amount = 0
        self.data_columns = []

        return self.extracted_cols

    def _fetch_and_parse(self, requested_row_amount, data_as='rows'):
        ''' See if this amount of data is available or a fetch from sqream is required
            -1 - fetch all available data. Used by fetchmany() '''

        if data_as == 'rows':
            while (requested_row_amount > len(self.parsed_rows)
                   or requested_row_amount == -1) and self.more_to_fetch:
                self.more_to_fetch = bool(self._fetch())  # _fetch() updates self.unparsed_row_amount

                self.parsed_rows.extend(zip(*self._parse_fetched_cols()))

    def execute(self, query, params=None):
        """Execute a statement. Parameters are not supported"""

        # self._verify_open()
        if params:

            log_and_raise(ProgrammingError, "Parametered queries not supported. \
                If this is an insert query, use executemany() with the data rows as the parameter")

        else:
            self._execute_sqream_statement(query)

        self._fill_description()
        self.rows_fetched = 0
        self.rows_returned = 0
        return self

    def executemany(self, query, rows_or_cols=None, data_as='rows', amount=None):
        ''' Execute a statement, including parametered data insert '''

        # self._verify_open()
        self.execute(query)

        if rows_or_cols is None:
            return self

        if data_as =='alchemy_flat_list':
            # Unflatten SQLalchemy data list
            row_len = len(self.column_list)
            rows_or_cols = [rows_or_cols[i: i +row_len] for i in range(0, len(rows_or_cols), row_len)]
            data_as = 'rows'

        if 'numpy' in repr(type(rows_or_cols[0])):
            data_as = 'numpy'

        # Network insert starts here if data was passed
        column_lengths = [len(row_or_col) for row_or_col in rows_or_cols]
        if column_lengths.count(column_lengths[0]) != len(column_lengths):
            log_and_raise(ProgrammingError,
                          "Incosistent data sequences passed for inserting. Please use rows/columns of consistent length"
                          )
        if data_as == 'rows':
            self.capacity = amount or len(rows_or_cols)
            self.cols = list(zip(*rows_or_cols))
        else:
            self.cols = rows_or_cols
            self.capacity = len(self.cols)

        # Slice a chunk of columns and pass to _send_columns()
        start_idx = 0
        while self.cols != [()]:
            col_chunk = [col[start_idx:start_idx + self.rows_per_flush] for col in self.cols]
            chunk_len = len(col_chunk[0])
            if chunk_len == 0:
                break
            self._send_columns(col_chunk, chunk_len)
            start_idx += self.rows_per_flush
            del col_chunk

            if logger.isEnabledFor(logging.INFO):
                logger.info(f'Sent {chunk_len} rows of data')

        self.close_stmt()

        return self

    def fetchmany(self, size=None, data_as='rows', fetchone=False):
        ''' Fetch an amount of result rows '''

        size = size or self.arraysize
        # self._verify_open()
        if self.statement_type not in (None, 'SELECT'):
            log_and_raise(ProgrammingError ,'No open statement while attempting fetch operation')

        if self.more_to_fetch is False:
            # All data from server for this select statement was fetched
            if len(self.parsed_rows) == 0:
                # Nothing
                return [] if not fetchone else None
        else:
            self._fetch_and_parse(size, data_as)

        # Get relevant part of parsed rows and reduce storage and counter
        if data_as == 'rows':
            res = self.parsed_rows[0:size if size != -1 else None]
            del self.parsed_rows[:size if size != -1 else None]

        if logger.isEnabledFor(logging.INFO):
            logger.info(f'Fetched {size} rows')

        return (res if res else []) if not fetchone else (res[0] if res else None)

    def fetchone(self, data_as='rows'):
        """Fetch one result row"""

        if data_as not in ('rows',):
            log_and_raise(ProgrammingError, "Bad argument to fetchone()")

        return self.fetchmany(1, data_as, fetchone=True)

    def fetchall(self, data_as='rows'):
        """Fetch all result rows"""

        if data_as not in ('rows',):
            log_and_raise(ProgrammingError, "Bad argument to fetchall()")

        return self.fetchmany(-1, data_as)

    # DB-API Do nothing (for now) methods
    # -----------------------------------

    def nextset(self):
        """No multiple result sets so currently always returns None"""
        log_and_raise(NotSupportedError, "Nextset is not supported")

    def setinputsizes(self, sizes):
        log_and_raise(NotSupportedError, "Setinputsizes is not supported")

    def setoutputsize(self, size, column=None):
        log_and_raise(NotSupportedError, "Setoutputsize is not supported")

    # def csv_to_table(self, csv_path, table_name, read=None, parse=None, convert=None, con=None, auto_infer=False,
    #                  delimiter="|"):
    #     ' Pyarrow CSV reader documentation: https://arrow.apache.org/docs/python/generated/pyarrow.csv.read_csv.html '
    #
    #     if not ARROW:
    #         return "Optional pyarrow dependency not found. To install: pip3 install pyarrow"
    #
    #     sqream_to_pa = {
    #         'ftBool':     pa.bool_(),
    #         'ftUByte':    pa.uint8(),
    #         'ftShort':    pa.int16(),
    #         'ftInt':      pa.int32(),
    #         'ftLong':     pa.int64(),
    #         'ftFloat':    pa.float32(),
    #         'ftDouble':   pa.float64(),
    #         'ftDate':     pa.timestamp('ns'),
    #         'ftDateTime': pa.timestamp('ns'),
    #         'ftVarchar':  pa.string(),
    #         'ftBlob':     pa.utf8(),
    #         'ftNumeric':  pa.decimal128(28, 11)
    #     }
    #
    #     start = time.time()
    #     # Get table metadata
    #     self.execute(f'select * from {table_name} where 1=0')
    #
    #     # Map column names to pyarrow types and set Arrow's CSV parameters
    #     sqream_col_types = [col_type[0] for col_type in self.col_type_tups]
    #     column_types = zip(self.col_names, [sqream_to_pa[col_type[0]] for col_type in self.col_type_tups])
    #     read = read or csv.ReadOptions(column_names=self.col_names)
    #     parse = parse or csv.ParseOptions(delimiter=delimiter)
    #     convert = convert or csv.ConvertOptions(column_types=None if auto_infer else column_types,
    #                                             null_values=["\\N"])
    #
    #     # Read CSV to in-memory arrow format
    #     csv_arrow = csv.read_csv(csv_path, read_options=read, parse_options=parse, convert_options=convert).\
    #         combine_chunks()
    #     num_chunks = len(csv_arrow[0].chunks)
    #     numpy_cols = []
    #
    #     # For each column, get the numpy representation for quick packing
    #     for col_type, col in zip(sqream_col_types, csv_arrow):
    #         # Only one chunk after combine_chunks()
    #         col = col.chunks[0]
    #         if col_type in  ('ftVarchar', 'ftBlob', 'ftDate', 'ftDateTime', 'ftNumeric'):
    #             col = col.to_pandas()
    #         else:
    #             col = col.to_numpy()
    #
    #         numpy_cols.append(col)
    #
    #     print(f'total loading csv: {time.time( ) -start}')
    #     start = time.time()
    #
    #     # Insert columns into SQream
    #     col_num = csv_arrow.shape[1]
    #     self.executemany(f'insert into {table_name} values ({"?, " *(col_num-1)}?)', numpy_cols)
    #     print(f'total inserting csv: {time.time( ) -start}')

    ## Closing

    def close_stmt(self):
        if self.open_statement:
            self.client.send_string('{"closeStatement": "closeStatement"}')
            self.open_statement = False

            if logger.isEnabledFor(logging.INFO):
                logger.info(f'Done executing statement {self.stmt_id} over connection {self.conn.connection_id}')

    def close(self, sock=None):
        self.close_stmt()
        sock = sock or self.s
        self.closed = True
        self.conn.close_connection()
        self.buffer.close()
        _end_ping_loop(self.ping_loop)
        self.cursors.pop(self.conn.connection_id)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()

    def __iter__(self):
        for item in self.fetchall():
            yield item


import logging

from pysqream.column_buffer import ColumnBuffer
from pysqream.SQSocket import SQSocket, Client
from ping import PingLoop, _end_ping_loop
from pysqream.globals import BUFFER_SIZE, FETCH_MANY_DEFAULT, CYTHON
from pysqream.logger import *
import json
import time
from queue import Queue, Empty
from struct import unpack
import socket
from pysqream.utils import NotSupportedError, ProgrammingError, InternalError, IntegrityError, OperationalError, DataError, \
    DatabaseError, InterfaceError, Warning, Error
from pysqream.casting import date_to_int as pydate_to_int, datetime_to_long as pydt_to_long, sq_date_to_py_date as date_to_py, sq_datetime_to_py_datetime as dt_to_py
from pysqream.cursor import Cursor


class Connection:
    ''' Connection class used to interact with SQream '''

    base_conn_open = [False]

    def __init__(self, ip, port, clustered, use_ssl=False, log=False, base_connection=True,
                 reconnect_attempts=3, reconnect_interval=10):

        self.buffer = ColumnBuffer(BUFFER_SIZE)  # flushing buffer every BUFFER_SIZE bytes
        self.version = None
        self.closed = False
        self.connect_to_socket = False
        self.connect_to_database = False
        self.orig_ip, self.orig_port, self.clustered, self.use_ssl = ip, port, clustered, use_ssl
        self.reconnect_attempts, self.reconnect_interval = reconnect_attempts, reconnect_interval
        self.base_connection = base_connection
        self.ping_loop = None
        self.client = None

        if self.base_connection:
            self.cursors = {}

        self._open_connection(clustered, use_ssl)

        self.unpack_q = Queue()
        if CYTHON:
            # To allow hot swapping for testing
            date_to_int, datetime_to_long, sq_date_to_py_date, sq_datetime_to_py_datetime = pydate_to_int, pydt_to_long, date_to_py, dt_to_py

        if log is not False:
            raise NotSupportedError("Logs per Connection is not supported yet")
            # start_logging(None if log is True else log)
        # Thread for unpacking fetched socket data
        # thread.start_new_thread(self._parse_fetched_cols, (self.unpack_q,))

    def __del__(self):
        try:
            logger.debug("Try to destroy open connections")
            self.close()
        except Exception as e:
            if "Trying to close a connection that's already closed" not in repr(e):
                log_and_raise(ProgrammingError, e)

    ## SQream mechanisms

    def _open_connection(self, clustered, use_ssl):
        ''' Get proper ip and port from picker if needed and connect socket. Used at __init__() '''
        if clustered is True:

            # Create non SSL socket for picker communication
            picker_socket = SQSocket(self.orig_ip, self.orig_port, False)
            self.client = Client(picker_socket)
            # Parse picker response to get ip and port
            # Read 4 bytes to find length of how much to read
            picker_socket.timeout(5)
            try:
                read_len = unpack('i', self.client.receive(4))[0]
            except socket.timeout:

                log_and_raise(ProgrammingError, f'Connected with clustered=True, but apparently not a server picker port')
            picker_socket.timeout(None)

            # Read the number of bytes, which is the IP in string format
            # Using a nonblocking socket in case clustered = True was passed but not connected to picker
            self.ip = self.client.receive(read_len)

            # Now read port
            self.port = unpack('i', self.client.receive(4))[0]
            picker_socket.close()
        else:
            self.ip, self.port = self.orig_ip, self.orig_port

        # Create socket and connect to actual SQreamd server
        self.s = SQSocket(self.ip, self.port, use_ssl)
        self.client = Client(self.s)
        self.connect_to_socket = True
        if self.base_connection:
            self.base_conn_open[0] = True

    def connect_database(self, database, username, password, service='sqream'):
        """Handle connection to database, with or without server picker"""

        self.database, self.username, self.password, self.service = database, username, password, service
        if self.connect_to_socket:
            res = self.client.send_string(
                f'{{"username":"{username}", "password":"{password}", "connectDatabase":"{database}", "service":"{service}"}}'
            )
            res = json.loads(res)
            try:
                self.connection_id = res['connectionId']
                if 'version' in res:
                    self.version = res['version']
            except KeyError as e:
                log_and_raise(ProgrammingError, f"Error connecting to database: {res['error']}")

            self.varchar_enc = res.get('varcharEncoding', 'ascii')

            if logger.isEnabledFor(logging.INFO):
                logger.info(f'Connection opened to database {database}. Connection ID: {self.connection_id}')
            self.connect_to_database = True

    def _attempt_reconnect(self):

        for attempt in range(self.reconnect_attempts):
            time.sleep(self.reconnect_interval)
            try:
                logger.info(f"attempt to connect numer {attempt}")
                self._open_connection(self.clustered, self.use_ssl)
                self.connect_database(self.database, self.username, self.password, self.service)
            except ConnectionRefusedError as e:
                print(f'Connection lost, retry attempt no. {attempt+1} failed. Original error:\n{e}')
            else:
                return

        # Attempts failed
        log_and_raise(ConnectionRefusedError, 'Reconnection attempts to sqreamd failed')

    def close_connection(self, sock=None):

        if self.closed:
            log_and_raise(ProgrammingError, f"Trying to close a connection that's already closed for database "
                                            f"{self.database} and Connection ID: {self.connection_id}")
        self.client.send_string('{"closeConnection":  "closeConnection"}')
        self.s.close()
        self.buffer.close()
        self.closed = True
        self.base_conn_open[0] = False if self.base_connection else True

        if logger.isEnabledFor(logging.INFO):
            logger.info(f'Connection closed to database {self.database}. Connection ID: {self.connection_id}')

    ## DB-API API
    #  ----------

    def _verify_open(self):

        if not self.base_conn_open[0]:
            log_and_raise(ProgrammingError, 'Connection has been closed')

        if self.closed:
            log_and_raise(ProgrammingError, 'Cursor has been closed')

    def cursor(self):
        ''' Return a new connection with the same parameters.
            We use a connection as the equivalent of a 'cursor' '''

        logger.debug("Create cursor")
        conn = Connection(
            self.orig_ip if self.clustered is True else self.ip,
            self.orig_port if self.clustered is True else self.port,
            self.clustered,
            self.use_ssl,
            base_connection=False
        )  # self is the calling connection instance, so cursor can trace back to pysqream
        conn.connect_database(self.database, self.username, self.password, self.service)

        self._verify_open()
        cur = Cursor(conn, self.cursors)
        self.cursors[cur.conn.connection_id] = cur
        return cur

    def commit(self):
        return
        # log_and_raise(NotSupportedError, "Commit is not supported")

    def rollback(self):
        return
        # log_and_raise(NotSupportedError, "Rollback is not supported")

    def close(self):
        if not self.connect_to_database:
            return

        if self.base_connection:
            for con_id, cursor in self.cursors.items():
                try:
                    cursor.close()
                except Exception as e:
                    logger.error(f"Can't close connection - {e} for Connection ID {con_id}")
                    raise Error(f"Can't close connection - {e} for Connection ID {con_id}")

        self.close_connection()


    # Internal Methods
    # ----------------
    ''' Include: __enter__(), __exit__() - for using "with", 
                 __iter__ for use in for-in clause  '''

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()